from setuptools import Extension
from setuptools import setup

import numpy as np
import os
from Cython.Build import cythonize

# skgrf project directory
top = os.path.dirname(os.path.abspath(__file__))

# include skgrf, grf, and numpy headers
# requires running buildpre.py to find src in this location
include_dirs = [
    top,
    os.path.join(top, "skgrf"),
    *[d[0] for d in os.walk(os.path.join(top, "skgrf", "grf", "src"))],
    *[d[0] for d in os.walk(os.path.join(top, "skgrf", "grf", "third_party"))],
    np.get_include(),
]


def find_ext_files(directory, ext="cpp", files=None):
    """Recursively find all Cython extension files.

    :param str directory: The directory in which to recursively crawl for .pyx files.
    :param list files: A list of files in which to append discovered .pyx files.
    """
    if files is None:
        files = []
    for file in os.listdir(directory):
        path = os.path.join(directory, file)
        if os.path.isfile(path) and path.endswith("." + ext):
            files.append(path)
        elif os.path.isdir(path):
            find_ext_files(path, ext, files)
    return files


def create_extension(module_name, additional_sources=None):
    """Create a setuptools build extension for a Cython extension file.

    :param str module_name: The name of the module
    :param list(str) additional_sources: A list of additional source filenames
    """
    path = module_name.replace(".", os.path.sep) + ".pyx"
    additional_sources = additional_sources or []
    for k in additional_sources:
        print(k)
    return Extension(
        module_name,
        sources=[path] + additional_sources,
        include_dirs=include_dirs,
        language="c++",
        extra_compile_args=["-std=c++11", "-Wall", "-O3"],
        extra_link_args=["-std=c++11", "-g"],
        # define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
    )


additional_sources = find_ext_files(os.path.join(top, "skgrf", "grf"), "cpp")
additional_sources = [
    f for f in additional_sources if "test" not in str(f).split("skgrf")[-1].lower()
]

ext_modules = [create_extension("skgrf.grf", additional_sources)]

setup(
    ext_modules=cythonize(
        ext_modules,
        gdb_debug=False,
        force=True,
        annotate=False,
        compiler_directives={"language_level": "3"},
    )
)


def build(setup_kwargs):
    setup_kwargs.update({"ext_modules": ext_modules})



from skgrf.ensemble.boosted_regressor import GRFBoostedForestRegressor
from skgrf.ensemble.causal_regressor import GRFForestCausalRegressor
from skgrf.ensemble.classifier import GRFForestClassifier
from skgrf.ensemble.instrumental_regressor import GRFForestInstrumentalRegressor
from skgrf.ensemble.local_linear_regressor import GRFForestLocalLinearRegressor
from skgrf.ensemble.quantile_regressor import GRFForestQuantileRegressor
from skgrf.ensemble.regressor import GRFForestRegressor
from skgrf.ensemble.survival import GRFForestSurvival



from skgrf.tree.causal_regressor import GRFTreeCausalRegressor
from skgrf.tree.classifier import GRFTreeClassifier
from skgrf.tree.instrumental_regressor import GRFTreeInstrumentalRegressor
from skgrf.tree.local_linear_regressor import GRFTreeLocalLinearRegressor
from skgrf.tree.quantile_regressor import GRFTreeQuantileRegressor
from skgrf.tree.regressor import GRFTreeRegressor
from skgrf.tree.survival import GRFTreeSurvival



import numpy as np
from sklearn.utils.validation import _check_sample_weight


def check_sample_weight(sample_weight, X):
    if sample_weight is not None:
        sample_weight = _check_sample_weight(sample_weight, X)
        use_sample_weight = True
        if np.array_equal(np.unique(sample_weight), np.array([1.0])):
            sample_weight = None
            use_sample_weight = False
    else:
        use_sample_weight = False

    return sample_weight, use_sample_weight



import multiprocessing
import numpy as np
from sklearn.base import BaseEstimator
from sklearn.tree._tree import csr_matrix
from sklearn.utils.validation import check_array
from sklearn.utils.validation import check_is_fitted

from skgrf import grf
from skgrf.base import GRFMixin


class BaseGRFForest(GRFMixin, BaseEstimator):
    def get_split_frequencies(self, max_depth=4):
        """Get the split frequencies of feature indexes at various depths.

        :param int max_depth: The maximum depth of splits to consider
        """
        check_is_fitted(self)
        return np.array(
            grf.compute_split_frequencies(self.grf_forest_cpp_, max_depth=max_depth)
        )

    def get_feature_importances(self, decay_exponent=2, max_depth=4):
        """Get the feature importances.

        :param int decay_exponent: Exponential decay of importance by split depth
        :param int max_depth: The maximum depth of splits to consider
        """
        sf = self.get_split_frequencies(max_depth=max_depth)
        sf = sf / np.maximum(1, np.sum(sf, axis=0))
        weight = np.arange(1.0, sf.shape[0] + 1) ** -decay_exponent
        return np.matmul(sf.T, weight / np.sum(weight))

    def get_kernel_weights(self, X, oob_prediction=False):
        """Get training sample weights for test data.

        Given a trained forest and test data,
        compute the kernel weights for each test point.

        Creates a sparse matrix in which the value at (i, j)
        gives the weight of training sample j for test sample i.
        Use ``oob_prediction=True`` if using training set.

        :param array2d X: input features
        :param bool oob_prediction: whether to calculate weights out of bag
        """
        check_is_fitted(self)
        X = check_array(X)
        self._check_n_features(X, reset=False)
        self._ensure_ptr()
        num_threads = self._get_num_threads()
        # this function doesn't accept 0, so we replace 0 with the real cpu count
        if num_threads == 0:
            num_threads = multiprocessing.cpu_count()
        samples, neighbors, weights = grf.compute_kernel_weights(
            self.grf_forest_cpp_,
            np.asfortranarray(X.astype("float64")),
            num_threads,
            oob_prediction,
        )
        return csr_matrix((weights, (samples, neighbors)))



import typing as t

import logging
import numpy as np
from sklearn.base import RegressorMixin
from sklearn.utils.validation import check_array
from sklearn.utils.validation import check_is_fitted

from skgrf import grf
from skgrf.tree.base import BaseGRFTree
from skgrf.utils.validation import check_sample_weight

if t.TYPE_CHECKING:  # pragma: no cover
    from skgrf.ensemble.instrumental_regressor import GRFForestInstrumentalRegressor

logger = logging.getLogger(__name__)


class GRFTreeInstrumentalRegressor(BaseGRFTree, RegressorMixin):
    r"""GRF Tree Instrumental regression implementation for sci-kit learn.

    Provides a sklearn tree instrumental regression to the GRF C++ library using Cython.

    :param bool equalize_cluster_weights: Weight the samples such that clusters have
        equally weight. If ``False``, larger clusters will have more weight. If
        ``True``, the number of samples drawn from each cluster is equal to the size of
        the smallest cluster. If ``True``, sample weights should not be passed on
        fitting.
    :param float sample_fraction: Fraction of samples used in each tree.
    :param int mtry: The number of features to split on each node. The default is
        ``sqrt(p) + 20`` where ``p`` is the number of features.
    :param int min_node_size: The minimum number of observations in each tree leaf.
    :param bool honesty: Use honest splitting (subsample splitting).
    :param float honesty_fraction: The fraction of data used for subsample splitting.
    :param bool honesty_prune_leaves: Prune estimation sample tree such that no leaves
        are empty. If ``False``, trees with empty leaves are skipped.
    :param float alpha: The maximum imbalance of a split.
    :param float imbalance_penalty: Penalty applied to imbalanced splits.
    :param double reduced_form_weight: Whether splits should be regularized towards a
        naive splitting criterion that ignores the instrument.
    :param bool stabilize_splits: Whether or not the instrument should be taken into
        account when determining the imbalance of a split.
    :param int n_jobs: The number of threads. Default is number of CPU cores. Only used
        for target estimation.
    :param int seed: Random seed value.

    :ivar int n_features_in\_: The number of features (columns) from the fit input
        ``X``.
    :ivar dict grf_forest\_: The returned result object from calling C++ grf.
    :ivar int mtry\_: The ``mtry`` value determined by validation.
    :ivar int outcome_index\_: The index of the grf train matrix holding the outcomes.
    :ivar list samples_per_cluster\_: The number of samples to train per cluster.
    :ivar list clusters\_: The cluster labels determined from the fit input ``cluster``.
    :ivar int n_clusters\_: The number of unique cluster labels from the fit input
        ``cluster``.
    :ivar str criterion: The criterion used for splitting: ``mse``
    """

    def __init__(
        self,
        equalize_cluster_weights=False,
        sample_fraction=0.5,
        mtry=None,
        min_node_size=5,
        honesty=True,
        honesty_fraction=0.5,
        honesty_prune_leaves=True,
        alpha=0.05,
        imbalance_penalty=0,
        reduced_form_weight=0,
        stabilize_splits=True,
        n_jobs=-1,
        seed=42,
    ):
        self.equalize_cluster_weights = equalize_cluster_weights
        self.sample_fraction = sample_fraction
        self.mtry = mtry
        self.min_node_size = min_node_size
        self.honesty = honesty
        self.honesty_fraction = honesty_fraction
        self.honesty_prune_leaves = honesty_prune_leaves
        self.alpha = alpha
        self.imbalance_penalty = imbalance_penalty
        self.reduced_form_weight = reduced_form_weight
        self.stabilize_splits = stabilize_splits
        self.n_jobs = n_jobs
        self.seed = seed

    @property
    def criterion(self):
        return "mse"

    @classmethod
    def from_forest(cls, forest: "GRFForestInstrumentalRegressor", idx: int):
        """Extract a tree from a forest.

        :param GRFLocalLinearRegressor forest: A trained GRFLocalLinearRegressor
            instance
        :param int idx: The tree index from the forest to extract.
        """
        # Even though we have a tree object, we keep the exact same dictionary structure
        # that exists in the forests, so that we can reuse the Cython entrypoints.
        # We also copy over some instance attributes from the trained forest.

        # params
        instance = cls(
            equalize_cluster_weights=forest.equalize_cluster_weights,
            sample_fraction=forest.sample_fraction,
            mtry=forest.mtry,
            min_node_size=forest.min_node_size,
            honesty=forest.honesty,
            honesty_fraction=forest.honesty_fraction,
            honesty_prune_leaves=forest.honesty_prune_leaves,
            alpha=forest.alpha,
            imbalance_penalty=forest.imbalance_penalty,
            reduced_form_weight=forest.reduced_form_weight,
            stabilize_splits=forest.stabilize_splits,
            n_jobs=forest.n_jobs,
            seed=forest.seed,
        )
        # forest
        grf_forest = {}
        for k, v in forest.grf_forest_.items():
            if isinstance(v, list):
                grf_forest[k] = [forest.grf_forest_[k][idx]]
            else:
                grf_forest[k] = v
        grf_forest["num_trees"] = 1
        instance.grf_forest_ = grf_forest
        instance._ensure_ptr()
        # vars
        instance.outcome_index_ = forest.outcome_index_
        instance.treatment_index_ = forest.treatment_index_
        instance.instrument_index_ = forest.instrument_index_
        instance.n_features_in_ = forest.n_features_in_
        instance.clusters_ = forest.clusters_
        instance.n_clusters_ = forest.n_clusters_
        instance.samples_per_cluster_ = forest.samples_per_cluster_
        instance.mtry_ = forest.mtry_
        instance.sample_weight_index_ = forest.sample_weight_index_
        return instance

    def fit(
        self,
        X,
        y,
        w,  # treatment
        z,  # instrument
        y_hat=None,
        w_hat=None,
        z_hat=None,
        sample_weight=None,
        cluster=None,
    ):
        """Fit the grf forest using training data.

        :param array2d X: training input features
        :param array1d y: training input targets
        :param array1d w: training input treatments
        :param array1d z: training input instruments
        :param array1d y_hat: estimated expected target responses
        :param array1d w_hat: estimated treatment propensities
        :param array1d z_hat: estimated instrument propensities
        :param array1d sample_weight: optional weights for input samples
        :param array1d cluster: optional cluster assignments for input samples
        """
        X, y = self._validate_data(X, y)
        self._check_num_samples(X)
        self._check_n_features(X, reset=True)

        self._check_sample_fraction()
        self._check_alpha()

        sample_weight, use_sample_weight = check_sample_weight(sample_weight, X)

        cluster = self._check_cluster(X=X, cluster=cluster)
        self.samples_per_cluster_ = self._check_equalize_cluster_weights(
            cluster=cluster, sample_weight=sample_weight
        )
        self.mtry_ = self._check_mtry(X=X)
        self._check_reduced_form_weight()

        if y_hat is None:
            logger.debug("estimating y_hat")
            y_hat = self._estimate_using_regression(
                X=X, y=y, sample_weight=sample_weight, cluster=cluster
            )

        if w_hat is None:
            logger.debug("estimating w_hat")
            w_hat = self._estimate_using_regression(
                X=X, y=w, sample_weight=sample_weight, cluster=cluster
            )

        # don't repeat calculations for causal
        if np.all(w == z):
            z_hat = w_hat

        if z_hat is None:
            logger.debug("estimating z_hat")
            z_hat = self._estimate_using_regression(
                X=X, y=z, sample_weight=sample_weight, cluster=cluster
            )

        y_centered = y - y_hat
        w_centered = w - w_hat
        z_centered = z - z_hat

        train_matrix = self._create_train_matrices(
            X=X,
            y=y_centered,
            sample_weight=sample_weight,
            treatment=w_centered,
            instrument=z_centered,
        )

        self.grf_forest_ = grf.instrumental_train(
            np.asfortranarray(train_matrix.astype("float64")),
            self.outcome_index_,
            self.treatment_index_,
            self.instrument_index_,
            self.sample_weight_index_,
            use_sample_weight,
            self.mtry_,
            1,  # num_trees
            self.min_node_size,
            self.sample_fraction,
            self.honesty,
            self.honesty_fraction,
            self.honesty_prune_leaves,
            1,  # ci_group_size
            self.reduced_form_weight,
            self.alpha,
            self.imbalance_penalty,
            self.stabilize_splits,
            cluster,
            self.samples_per_cluster_,
            False,  # compute_oob_predictions,
            1,  # num_threads,
            self.seed,
        )
        self._ensure_ptr()
        sample_weight = sample_weight if sample_weight is not None else np.ones(len(X))
        self._set_node_values(y, sample_weight)
        self._set_n_classes()
        return self

    def predict(self, X):
        """Predict regression target for X.

        :param array2d X: prediction input features
        """
        return np.atleast_1d(np.squeeze(np.array(self._predict(X)["predictions"])))

    def _predict(self, X, estimate_variance=False):
        check_is_fitted(self)
        X = check_array(X)
        self._check_n_features(X, reset=False)
        self._ensure_ptr()

        result = grf.instrumental_predict(
            self.grf_forest_cpp_,
            np.asfortranarray([[]]),  # train_matrix
            self.outcome_index_,
            self.treatment_index_,
            self.instrument_index_,
            np.asfortranarray(X.astype("float64")),  # test_matrix
            self._get_num_threads(),
            estimate_variance,
        )
        return result

    def _estimate_using_regression(self, X, y, sample_weight=None, cluster=None):
        """Generate target estimates using a regression.

        In the R package, they perform forest tuning here. For now, we just perform
        a single regression without tuning. We also don't expose any of the forest
        parametrization for this process in the estimator.

        # TODO consider implementing tuning, exposing parameters.
        """
        train_matrix = self._create_train_matrices(
            X=X,
            y=y,
            sample_weight=sample_weight,
        )
        sample_weight, use_sample_weight = check_sample_weight(sample_weight, X)
        n_estimators = 50
        regression_forest = grf.regression_train(
            np.asfortranarray(train_matrix.astype("float64")),
            self.outcome_index_,
            self.sample_weight_index_,
            use_sample_weight,
            self.mtry_,
            n_estimators,  # num_trees
            5,  # min_node_size
            self.sample_fraction,
            True,  # honesty
            0.5,  # honesty_fraction
            self.honesty_prune_leaves,
            1,  # ci_group_size
            self.alpha,
            self.imbalance_penalty,
            cluster,
            self.samples_per_cluster_,
            True,  # compute_oob_predictions,
            self._get_num_threads(),  # num_threads,
            self.seed,
        )
        return np.atleast_1d(np.squeeze(np.array(regression_forest["predictions"])))

    def _more_tags(self):  # pragma: no cover
        return {
            "requires_y": True,
            "_xfail_checks": {
                "check_sample_weights_invariance": "zero sample_weight is not equivalent to removing samples",
            },
        }



import typing as t

import numpy as np
from sklearn.base import ClassifierMixin
from sklearn.utils.multiclass import check_classification_targets
from sklearn.utils.validation import check_array
from sklearn.utils.validation import check_is_fitted

from skgrf import grf
from skgrf.tree.base import BaseGRFTree
from skgrf.utils.validation import check_sample_weight

if t.TYPE_CHECKING:  # pragma: no cover
    from skgrf.ensemble.classifier import GRFForestClassifier


class GRFTreeClassifier(BaseGRFTree, ClassifierMixin):
    r"""GRF Tree Classification implementation for sci-kit learn.

    Provides a sklearn tree classifier interface to the GRF C++ library using Cython.

    :param bool equalize_cluster_weights: Weight the samples such that clusters have
        equally weight. If ``False``, larger clusters will have more weight. If
        ``True``, the number of samples drawn from each cluster is equal to the size of
        the smallest cluster. If ``True``, sample weights should not be passed on
        fitting.
    :param float sample_fraction: Fraction of samples used in each tree.
    :param int mtry: The number of features to split on each node. The default is
        ``sqrt(p) + 20`` where ``p`` is the number of features.
    :param int min_node_size: The minimum number of observations in each tree leaf.
    :param bool honesty: Use honest splitting (subsample splitting).
    :param float honesty_fraction: The fraction of data used for subsample splitting.
    :param bool honesty_prune_leaves: Prune estimation sample tree such that no leaves
        are empty. If ``False``, trees with empty leaves are skipped.
    :param float alpha: The maximum imbalance of a split.
    :param float imbalance_penalty: Penalty applied to imbalanced splits.
    :param int seed: Random seed value.

    :ivar int n_features_in\_: The number of features (columns) from the fit input
        ``X``.
    :ivar dict grf_forest\_: The returned result object from calling C++ grf.
    :ivar int mtry\_: The ``mtry`` value determined by validation.
    :ivar int outcome_index\_: The index of the grf train matrix holding the outcomes.
    :ivar list samples_per_cluster\_: The number of samples to train per cluster.
    :ivar list clusters\_: The cluster labels determined from the fit input ``cluster``.
    :ivar int n_clusters\_: The number of unique cluster labels from the fit input
        ``cluster``.
    :ivar list classes\_: The class labels determined from the fit input ``y``.
    :ivar int n_classes\_: The number of unique class labels from the fit input
        ``y``.
    :ivar str criterion: The criterion used for splitting: ``gini``
    """

    def __init__(
        self,
        equalize_cluster_weights=False,
        sample_fraction=0.5,
        mtry=None,
        min_node_size=5,
        honesty=True,
        honesty_fraction=0.5,
        honesty_prune_leaves=True,
        alpha=0.05,
        imbalance_penalty=0.0,
        seed=42,
    ):
        self.equalize_cluster_weights = equalize_cluster_weights
        self.sample_fraction = sample_fraction
        self.mtry = mtry
        self.min_node_size = min_node_size
        self.honesty = honesty
        self.honesty_fraction = honesty_fraction
        self.honesty_prune_leaves = honesty_prune_leaves
        self.alpha = alpha
        self.imbalance_penalty = imbalance_penalty
        self.seed = seed

    @property
    def criterion(self):
        return "gini"

    @classmethod
    def from_forest(cls, forest: "GRFForestClassifier", idx: int):
        """Extract a tree from a forest.

        :param GRFForestClassifier forest: A trained GRFClassifier instance
        :param int idx: The tree index from the forest to extract.
        """
        # Even though we have a tree object, we keep the exact same dictionary structure
        # that exists in the forests, so that we can reuse the Cython entrypoints.
        # We also copy over some instance attributes from the trained forest.

        # params
        instance = cls(
            equalize_cluster_weights=forest.equalize_cluster_weights,
            sample_fraction=forest.sample_fraction,
            mtry=forest.mtry,
            min_node_size=forest.min_node_size,
            honesty=forest.honesty,
            honesty_fraction=forest.honesty_fraction,
            honesty_prune_leaves=forest.honesty_prune_leaves,
            alpha=forest.alpha,
            imbalance_penalty=forest.imbalance_penalty,
            seed=forest.seed,
        )
        # forest
        grf_forest = {}
        for k, v in forest.grf_forest_.items():
            if isinstance(v, list):
                grf_forest[k] = [forest.grf_forest_[k][idx]]
            else:
                grf_forest[k] = v
        grf_forest["num_trees"] = 1
        instance.grf_forest_ = grf_forest
        instance._ensure_ptr()
        # vars
        instance.outcome_index_ = forest.outcome_index_
        instance.n_features_in_ = forest.n_features_in_
        instance.clusters_ = forest.clusters_
        instance.n_clusters_ = forest.n_clusters_
        instance.classes_ = forest.classes_
        instance.n_classes_ = forest.n_classes_
        instance.samples_per_cluster_ = forest.samples_per_cluster_
        instance.mtry_ = forest.mtry_
        instance.sample_weight_index_ = forest.sample_weight_index_
        return instance

    def fit(
        self, X, y, sample_weight=None, cluster=None, compute_oob_predictions=False
    ):
        """Fit the grf forest using training data.

        :param array2d X: training input features
        :param array1d y: training input targets
        :param array1d sample_weight: optional weights for input samples
        :param array1d cluster: optional cluster assignments for input samples
        """
        X, y = self._validate_data(X, y)
        check_classification_targets(y)
        self._check_num_samples(X)
        self._check_n_features(X, reset=True)

        self._check_sample_fraction(oob=compute_oob_predictions)
        self._check_alpha()

        sample_weight, use_sample_weight = check_sample_weight(sample_weight, X)

        cluster_ = self._check_cluster(X=X, cluster=cluster)
        self.samples_per_cluster_ = self._check_equalize_cluster_weights(
            cluster=cluster_, sample_weight=sample_weight
        )
        self.mtry_ = self._check_mtry(X=X)

        # Map classes to indices
        y = np.copy(y)
        self.classes_, y = np.unique(y, return_inverse=True)
        self.n_classes_ = len(self.classes_)

        train_matrix = self._create_train_matrices(
            X=X, y=y, sample_weight=sample_weight
        )

        self.grf_forest_ = grf.probability_train(
            np.asfortranarray(train_matrix.astype("float64")),
            self.outcome_index_,
            self.sample_weight_index_,
            use_sample_weight,
            self.n_classes_,
            self.mtry_,
            1,  # num_trees
            self.min_node_size,
            self.sample_fraction,
            self.honesty,
            self.honesty_fraction,
            self.honesty_prune_leaves,
            1,  # ci_group_size
            self.alpha,
            self.imbalance_penalty,
            cluster_,
            self.samples_per_cluster_,
            compute_oob_predictions,
            1,  # num_threads
            self.seed,
        )
        self._ensure_ptr()
        sample_weight = sample_weight if sample_weight is not None else np.ones(len(X))
        self._set_node_values(y, sample_weight)
        self._set_n_classes()
        return self

    def predict(self, X):
        """Predict classes from X.

        :param array2d X: prediction input features
        """
        probas = self.predict_proba(X)
        return self.classes_.take(np.argmax(probas, axis=1), axis=0)

    def predict_proba(self, X):
        """Predict probabilities for classes from X.

        :param array2d X: prediction input features
        """
        result = self._predict(X)
        predictions = np.atleast_2d(np.array(result["predictions"]))
        return predictions

    def predict_log_proba(self, X):
        """Predict log probabilities for classes from X.

        :param array2d X: prediction input features
        """
        proba = self.predict_proba(X)
        return np.log(proba)

    def _predict(self, X, estimate_variance=False):
        check_is_fitted(self)
        X = check_array(X)
        self._check_n_features(X, reset=False)
        self._ensure_ptr()

        result = grf.probability_predict(
            self.grf_forest_cpp_,
            np.asfortranarray([[]]),  # train_matrix
            self.outcome_index_,
            self.n_classes_,
            np.asfortranarray(X.astype("float64")),  # test_matrix
            1,  # num_threads
            estimate_variance,
        )
        return result

    def _more_tags(self):
        return {
            "_xfail_checks": {
                "check_sample_weights_invariance": "zero sample_weight is not equivalent to removing samples",
            },
        }



import numpy as np
from sklearn.base import RegressorMixin
from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_array
from sklearn.utils.validation import check_is_fitted

from skgrf import grf
from skgrf.ensemble.base import BaseGRFForest
from skgrf.tree.quantile_regressor import GRFTreeQuantileRegressor


class GRFForestQuantileRegressor(BaseGRFForest, RegressorMixin):
    r"""GRF Quantile Regression implementation for sci-kit learn.

    Provides a sklearn quantile regressor interface to the GRF C++ library using Cython.

    .. warning::

        Because the training dataset is required for prediction, the training dataset
        is recorded onto the estimator instance. This means that serializing this
        estimator will result in a file at least as large as the serialized training
        dataset.

    :param int n_estimators: The number of tree regressors to train
    :param list(float) quantiles: A list of quantiles on which to predict.
    :param bool regression_splitting: Use regression splits instead of splitting
        specially for quantiles.
    :param bool equalize_cluster_weights: Weight the samples such that clusters have
        equally weight. If ``False``, larger clusters will have more weight. If
        ``True``, the number of samples drawn from each cluster is equal to the size of
        the smallest cluster. If ``True``, sample weights should not be passed on
        fitting.
    :param float sample_fraction: Fraction of samples used in each tree.
    :param int mtry: The number of features to split on each node. The default is
        ``sqrt(p) + 20`` where ``p`` is the number of features.
    :param int min_node_size: The minimum number of observations in each tree leaf.
    :param bool honesty: Use honest splitting (subsample splitting).
    :param float honesty_fraction: The fraction of data used for subsample splitting.
    :param bool honesty_prune_leaves: Prune estimation sample tree such that no leaves
        are empty. If ``False``, trees with empty leaves are skipped.
    :param float alpha: The maximum imbalance of a split.
    :param float imbalance_penalty: Penalty applied to imbalanced splits.
    :param int n_jobs: The number of threads. Default is number of CPU cores.
    :param int seed: Random seed value.
    :param bool enable_tree_details: When ``True``, perform additional calculations
        for detailing the underlying decision trees. Must be enabled for ``estimators_``
        and ``get_estimator`` to work. Very slow.

    :ivar list estimators\_: A list of tree objects from the forest.
    :ivar int n_features_in\_: The number of features (columns) from the fit input
        ``X``.
    :ivar dict grf_forest\_: The returned result object from calling C++ grf.
    :ivar int mtry\_: The ``mtry`` value determined by validation.
    :ivar int outcome_index\_: The index of the grf train matrix holding the outcomes.
    :ivar list samples_per_cluster\_: The number of samples to train per cluster.
    :ivar list clusters\_: The cluster labels determined from the fit input ``cluster``.
    :ivar int n_clusters\_: The number of unique cluster labels from the fit input
        ``cluster``.
    :ivar array2d train\_: The ``X,y`` concatenated train matrix passed to grf.
    :ivar str criterion: The criterion used for splitting: ``gini``
    """

    def __init__(
        self,
        n_estimators=100,
        quantiles=None,
        regression_splitting=False,
        equalize_cluster_weights=False,
        sample_fraction=0.5,
        mtry=None,
        min_node_size=5,
        honesty=True,
        honesty_fraction=0.5,
        honesty_prune_leaves=True,
        alpha=0.05,
        imbalance_penalty=0,
        n_jobs=-1,
        seed=42,
        enable_tree_details=False,
    ):
        self.n_estimators = n_estimators
        self.quantiles = quantiles
        self.regression_splitting = regression_splitting
        self.equalize_cluster_weights = equalize_cluster_weights
        self.sample_fraction = sample_fraction
        self.mtry = mtry
        self.min_node_size = min_node_size
        self.honesty = honesty
        self.honesty_fraction = honesty_fraction
        self.honesty_prune_leaves = honesty_prune_leaves
        self.alpha = alpha
        self.imbalance_penalty = imbalance_penalty
        self.n_jobs = n_jobs
        self.seed = seed
        self.enable_tree_details = enable_tree_details

    @property
    def criterion(self):
        return "gini"

    @property
    def estimators_(self):
        try:
            check_is_fitted(self)
        except NotFittedError:
            raise AttributeError(
                f"{self.__class__.__name__} object has no attribute 'estimators_'"
            ) from None
        if not self.enable_tree_details:
            raise ValueError("enable_tree_details must be True prior to training")
        return [
            GRFTreeQuantileRegressor.from_forest(self, idx=idx)
            for idx in range(self.n_estimators)
        ]

    def get_estimator(self, idx):
        """Extract a single estimator tree from the forest.

        :param int idx: The index of the tree to extract.
        """
        check_is_fitted(self)
        if not self.enable_tree_details:
            raise ValueError("enable_tree_details must be True prior to training")
        return GRFTreeQuantileRegressor.from_forest(self, idx=idx)

    def fit(self, X, y, cluster=None):
        """Fit the grf quantile forest using training data.

        :param array2d X: training input features
        :param array1d y: training input targets
        :param array1d cluster: optional cluster assignments for input samples
        """
        if self.quantiles is None:
            raise ValueError("quantiles must be set")

        X, y = self._validate_data(X, y)
        self._check_num_samples(X)
        self._check_n_features(X, reset=True)

        self._check_sample_fraction()
        self._check_alpha()

        cluster = self._check_cluster(X=X, cluster=cluster)
        self.samples_per_cluster_ = self._check_equalize_cluster_weights(
            cluster=cluster, sample_weight=None
        )
        self.mtry_ = self._check_mtry(X=X)

        train_matrix = self._create_train_matrices(X, y)
        self.train_ = train_matrix

        self.grf_forest_ = grf.quantile_train(
            self.quantiles,
            self.regression_splitting,
            np.asfortranarray(train_matrix.astype("float64")),
            self.outcome_index_,
            self.mtry_,
            self.n_estimators,  # num_trees
            self.min_node_size,
            self.sample_fraction,
            self.honesty,
            self.honesty_fraction,
            self.honesty_prune_leaves,
            1,  # ci_group_size,
            self.alpha,
            self.imbalance_penalty,
            cluster,
            self.samples_per_cluster_,
            False,  # compute_oob_predictions,
            self._get_num_threads(),  # num_threads
            self.seed,
        )
        self._ensure_ptr()

        if self.enable_tree_details:
            sample_weight = np.ones(len(X))
            self._set_node_values(y, sample_weight)
            self._set_n_classes()

        return self

    def predict(self, X):
        """Predict quantile regression target(s) for X.

        :param array2d X: prediction input features
        """
        return np.atleast_1d(np.squeeze(np.array(self._predict(X)["predictions"])))

    def _predict(self, X):
        check_is_fitted(self)
        X = check_array(X)
        self._check_n_features(X, reset=False)
        self._ensure_ptr()

        result = grf.quantile_predict(
            self.grf_forest_cpp_,
            self.quantiles,
            np.asfortranarray(self.train_.astype("float64")),
            self.outcome_index_,
            np.asfortranarray(X.astype("float64")),  # test_matrix
            self._get_num_threads(),  # num_threads
        )
        return result

    def _more_tags(self):
        return {"requires_y": True, "poor_score": True}



import cython
import numpy as np
import sys

cimport numpy as np
from cython.operator cimport dereference as deref
from cython.operator cimport postincrement as postinc
from libcpp.memory cimport bool
from libcpp.memory cimport unique_ptr
from libcpp.unordered_map cimport unordered_map
from libcpp.utility cimport move
from libcpp.vector cimport vector

from skgrf cimport grf_


cdef class Data:
    """Cython wrapper for Data C++ class.

    This wraps the Data class in C++, which encapsulates training data passed to the
    random forest classes. It allows us to pass numpy arrays as a grf-compatible
    Data object.
    """
    cdef unique_ptr[grf_.Data] c_data

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def __cinit__(self,
        np.ndarray[double, ndim=2, mode="fortran"] data not None,
    ):
        cdef size_t num_rows = np.PyArray_DIMS(data)[0]
        cdef size_t num_cols = np.PyArray_DIMS(data)[1]
        self.c_data.reset(
            new grf_.Data(
                &data[0, 0],
                num_rows,
                num_cols,
            )
        )

    def get(self, size_t row, size_t col):
        return deref(self.c_data).get(row, col)


cdef class GRFForest:
    """Cython wrapper for forest pointer.

    Holds a reference to the forest so that we don't have to
    deserialize on each predict call.
    """
    cdef grf_.Forest* forest

    def __cinit__(self, dict forest_object):
        self.forest = deserialize_forest(forest_object)


cpdef regression_train(
    np.ndarray[double, ndim=2, mode="fortran"] train_matrix,
    size_t outcome_index,
    size_t sample_weight_index,
    bool use_sample_weights,
    unsigned int mtry,
    unsigned int num_trees,
    unsigned int min_node_size,
    double sample_fraction,
    bool honesty,
    double honesty_fraction,
    bool honesty_prune_leaves,
    size_t ci_group_size,
    double alpha,
    double imbalance_penalty,
    vector[size_t] clusters,
    unsigned int samples_per_cluster,
    bool compute_oob_predictions,
    unsigned int num_threads,
    unsigned int seed,
):
    cdef grf_.ForestOptions* options
    cdef vector[grf_.Prediction] predictions

    trainer = new grf_.ForestTrainer(grf_.regression_trainer())
    data = Data(train_matrix)
    deref(data.c_data).set_outcome_index(outcome_index)

    if use_sample_weights:
        deref(data.c_data).set_weight_index(sample_weight_index - 1)

    options = new grf_.ForestOptions(
        num_trees,
        ci_group_size,
        sample_fraction,
        mtry,
        min_node_size,
        honesty,
        honesty_fraction,
        honesty_prune_leaves,
        alpha,
        imbalance_penalty,
        num_threads,
        seed,
        clusters,
        samples_per_cluster,
    )

    forest = new grf_.Forest(trainer.train(deref(data.c_data), deref(options)))

    if compute_oob_predictions:
        predictor = new grf_.ForestPredictor(grf_.regression_predictor(num_threads))
        predictions = predictor.predict_oob(deref(forest), deref(data.c_data), False)

    return create_forest_object(forest, predictions)


cpdef regression_predict(
    GRFForest forest_wrapper,
    np.ndarray[double, ndim=2, mode="fortran"] train_matrix,
    size_t outcome_index,
    np.ndarray[double, ndim=2, mode="fortran"] test_matrix,
    unsigned int num_threads,
    bool estimate_variance,
):
    cdef vector[grf_.Prediction] predictions

    predictor = new grf_.ForestPredictor(grf_.regression_predictor(num_threads))
    train_data = Data(train_matrix)
    deref(train_data.c_data).set_outcome_index(outcome_index)

    test_data = Data(test_matrix)

    predictions = predictor.predict(
        deref(forest_wrapper.forest),
        deref(train_data.c_data),
        deref(test_data.c_data),
        estimate_variance,
    )

    return create_prediction_object(predictions)


cpdef quantile_train(
    vector[double] quantiles,
    bool regression_splitting,
    np.ndarray[double, ndim=2, mode="fortran"] train_matrix,
    size_t outcome_index,
    unsigned int mtry,
    unsigned int num_trees,
    int min_node_size,
    double sample_fraction,
    bool honesty,
    double honesty_fraction,
    bool honesty_prune_leaves,
    size_t ci_group_size,
    double alpha,
    double imbalance_penalty,
    vector[size_t] clusters,
    unsigned int samples_per_cluster,
    bool compute_oob_predictions,
    int num_threads,
    unsigned int seed,
):
    cdef grf_.ForestOptions* options
    cdef vector[grf_.Prediction] predictions

    if regression_splitting:
        trainer = new grf_.ForestTrainer(grf_.regression_trainer())
    else:
        trainer = new grf_.ForestTrainer(grf_.quantile_trainer(quantiles))

    data = Data(train_matrix)
    deref(data.c_data).set_outcome_index(outcome_index)

    options = new grf_.ForestOptions(
        num_trees,
        ci_group_size,
        sample_fraction,
        mtry,
        min_node_size,
        honesty,
        honesty_fraction,
        honesty_prune_leaves,
        alpha,
        imbalance_penalty,
        num_threads,
        seed,
        clusters,
        samples_per_cluster,
    )

    forest = new grf_.Forest(trainer.train(deref(data.c_data), deref(options)))

    if compute_oob_predictions:
        predictor = new grf_.ForestPredictor(grf_.quantile_predictor(num_threads, quantiles))

    return create_forest_object(forest, predictions)


cpdef quantile_predict(
    GRFForest forest_wrapper,
    vector[double] quantiles,
    np.ndarray[double, ndim=2, mode="fortran"] train_matrix,
    size_t outcome_index,
    np.ndarray[double, ndim=2, mode="fortran"] test_matrix,
    unsigned int num_threads,
):
    cdef vector[grf_.Prediction] predictions

    predictor = new grf_.ForestPredictor(grf_.quantile_predictor(num_threads, quantiles))
    train_data = Data(train_matrix)
    deref(train_data.c_data).set_outcome_index(outcome_index)

    test_data = Data(test_matrix)

    predictions = predictor.predict(
        deref(forest_wrapper.forest),
        deref(train_data.c_data),
        deref(test_data.c_data),
        False,  # estimate_variance
    )

    return create_prediction_object(predictions)


cpdef survival_train(
    np.ndarray[double, ndim=2, mode="fortran"] train_matrix,
    size_t outcome_index,
    size_t censor_index,
    size_t sample_weight_index,
    bool use_sample_weights,
    unsigned int mtry,
    unsigned int num_trees,
    unsigned int min_node_size,
    double sample_fraction,
    bool honesty,
    double honesty_fraction,
    bool honesty_prune_leaves,
    double alpha,
    size_t num_failures,
    vector[size_t] clusters,
    unsigned int samples_per_cluster,
    bool compute_oob_predictions,
    int num_threads,
    unsigned int seed,
):
    cdef grf_.ForestOptions* options
    cdef vector[grf_.Prediction] predictions

    trainer = new grf_.ForestTrainer(grf_.survival_trainer())

    data = Data(train_matrix)
    deref(data.c_data).set_outcome_index(outcome_index)
    deref(data.c_data).set_censor_index(censor_index)
    if use_sample_weights:
        deref(data.c_data).set_weight_index(sample_weight_index)

    options = new grf_.ForestOptions(
        num_trees,
        1,  # ci_group_size
        sample_fraction,
        mtry,
        min_node_size,
        honesty,
        honesty_fraction,
        honesty_prune_leaves,
        alpha,
        0,  # imbalance_penalty
        num_threads,
        seed,
        clusters,
        samples_per_cluster,
    )

    forest = new grf_.Forest(trainer.train(deref(data.c_data), deref(options)))

    if compute_oob_predictions:
        predictor = new grf_.ForestPredictor(grf_.survival_predictor(num_threads, num_failures, 0))

    return create_forest_object(forest, predictions)


cpdef survival_predict(
    GRFForest forest_wrapper,
    np.ndarray[double, ndim=2, mode="fortran"] train_matrix,
    size_t outcome_index,
    size_t censor_index,
    size_t sample_weight_index,
    bool use_sample_weights,
    np.ndarray[double, ndim=2, mode="fortran"] test_matrix,
    unsigned int num_threads,
    size_t num_failures,
):
    cdef vector[grf_.Prediction] predictions

    predictor = new grf_.ForestPredictor(grf_.survival_predictor(num_threads, num_failures, 0))
    train_data = Data(train_matrix)
    deref(train_data.c_data).set_outcome_index(outcome_index)
    deref(train_data.c_data).set_censor_index(censor_index)
    if use_sample_weights:
        deref(train_data.c_data).set_weight_index(sample_weight_index)

    test_data = Data(test_matrix)

    predictions = predictor.predict(
        deref(forest_wrapper.forest),
        deref(train_data.c_data),
        deref(test_data.c_data),
        False,  # estimate_variance
    )

    return create_prediction_object(predictions)


cpdef ll_regression_train(
    np.ndarray[double, ndim=2, mode="fortran"] train_matrix,
    size_t outcome_index,
    size_t sample_weight_index,
    double ll_split_lambda,
    bool ll_split_weight_penalty,
    vector[size_t] ll_split_variables,
    size_t ll_split_cutoff,
    vector[double] overall_beta,
    bool use_sample_weights,
    unsigned int mtry,
    unsigned int num_trees,
    unsigned int min_node_size,
    double sample_fraction,
    bool honesty,
    double honesty_fraction,
    bool honesty_prune_leaves,
    size_t ci_group_size,
    double alpha,
    double imbalance_penalty,
    vector[size_t] clusters,
    unsigned int samples_per_cluster,
    unsigned int num_threads,
    unsigned int seed,
):
    cdef grf_.ForestOptions* options
    cdef vector[grf_.Prediction] predictions

    trainer = new grf_.ForestTrainer(
        grf_.ll_regression_trainer(
            ll_split_lambda,
            ll_split_weight_penalty,
            overall_beta,
            ll_split_cutoff,
            ll_split_variables,
        )
    )

    data = Data(train_matrix)
    deref(data.c_data).set_outcome_index(outcome_index)
    if use_sample_weights:
        deref(data.c_data).set_weight_index(sample_weight_index)

    options = new grf_.ForestOptions(
        num_trees,
        ci_group_size,
        sample_fraction,
        mtry,
        min_node_size,
        honesty,
        honesty_fraction,
        honesty_prune_leaves,
        alpha,
        imbalance_penalty,
        num_threads,
        seed,
        clusters,
        samples_per_cluster,
    )

    forest = new grf_.Forest(trainer.train(deref(data.c_data), deref(options)))

    return create_forest_object(forest, predictions)


cpdef ll_regression_predict(
    GRFForest forest_wrapper,
    np.ndarray[double, ndim=2, mode="fortran"] train_matrix,
    size_t outcome_index,
    np.ndarray[double, ndim=2, mode="fortran"] test_matrix,
    vector[double] ll_lambda,
    bool ll_weight_penalty,
    vector[size_t] linear_correction_variables,
    unsigned int num_threads,
    bool estimate_variance,
):
    cdef vector[grf_.Prediction] predictions

    predictor = new grf_.ForestPredictor(
        grf_.ll_regression_predictor(
            num_threads,
            ll_lambda,
            ll_weight_penalty,
            linear_correction_variables,
        )
    )
    train_data = Data(train_matrix)
    deref(train_data.c_data).set_outcome_index(outcome_index)

    test_data = Data(test_matrix)

    predictions = predictor.predict(
        deref(forest_wrapper.forest),
        deref(train_data.c_data),
        deref(test_data.c_data),
        estimate_variance,
    )

    return create_prediction_object(predictions)


cpdef instrumental_train(
    np.ndarray[double, ndim=2, mode="fortran"] train_matrix,
    size_t outcome_index,
    size_t treatment_index,
    size_t instrument_index,
    size_t sample_weight_index,
    bool use_sample_weights,
    unsigned int mtry,
    unsigned int num_trees,
    unsigned int min_node_size,
    double sample_fraction,
    bool honesty,
    double honesty_fraction,
    bool honesty_prune_leaves,
    size_t ci_group_size,
    double reduced_form_weight,
    double alpha,
    double imbalance_penalty,
    bool stabilize_splits,
    vector[size_t] clusters,
    unsigned int samples_per_cluster,
    bool compute_oob_predictions,
    unsigned int num_threads,
    unsigned int seed,
):
    cdef grf_.ForestOptions* options
    cdef vector[grf_.Prediction] predictions

    trainer = new grf_.ForestTrainer(
        grf_.instrumental_trainer(
            reduced_form_weight, stabilize_splits,
        )
    )

    data = Data(train_matrix)
    deref(data.c_data).set_outcome_index(outcome_index)
    deref(data.c_data).set_treatment_index(treatment_index)
    deref(data.c_data).set_instrument_index(instrument_index)
    if use_sample_weights:
        deref(data.c_data).set_weight_index(sample_weight_index)

    options = new grf_.ForestOptions(
        num_trees,
        ci_group_size,
        sample_fraction,
        mtry,
        min_node_size,
        honesty,
        honesty_fraction,
        honesty_prune_leaves,
        alpha,
        imbalance_penalty,
        num_threads,
        seed,
        clusters,
        samples_per_cluster,
    )

    forest = new grf_.Forest(trainer.train(deref(data.c_data), deref(options)))

    if compute_oob_predictions:
        predictor = new grf_.ForestPredictor(grf_.instrumental_predictor(num_threads))

    return create_forest_object(forest, predictions)


cpdef instrumental_predict(
    GRFForest forest_wrapper,
    np.ndarray[double, ndim=2, mode="fortran"] train_matrix,
    size_t outcome_index,
    size_t treatment_index,
    size_t instrument_index,
    np.ndarray[double, ndim=2, mode="fortran"] test_matrix,
    unsigned int num_threads,
    bool estimate_variance,
):
    cdef vector[grf_.Prediction] predictions

    predictor = new grf_.ForestPredictor(
        grf_.instrumental_predictor(
            num_threads,
        )
    )
    train_data = Data(train_matrix)
    deref(train_data.c_data).set_outcome_index(outcome_index)
    deref(train_data.c_data).set_treatment_index(treatment_index)
    deref(train_data.c_data).set_instrument_index(instrument_index)

    test_data = Data(test_matrix)

    predictions = predictor.predict(
        deref(forest_wrapper.forest),
        deref(train_data.c_data),
        deref(test_data.c_data),
        estimate_variance,
    )

    return create_prediction_object(predictions)


cpdef causal_train(
    np.ndarray[double, ndim=2, mode="fortran"] train_matrix,
    size_t outcome_index,
    size_t treatment_index,
    size_t sample_weight_index,
    bool use_sample_weights,
    unsigned int mtry,
    unsigned int num_trees,
    unsigned int min_node_size,
    double sample_fraction,
    bool honesty,
    double honesty_fraction,
    bool honesty_prune_leaves,
    size_t ci_group_size,
    double reduced_form_weight,
    double alpha,
    double imbalance_penalty,
    bool stabilize_splits,
    vector[size_t] clusters,
    unsigned int samples_per_cluster,
    bool compute_oob_predictions,
    unsigned int num_threads,
    unsigned int seed,
):
    cdef grf_.ForestOptions* options
    cdef vector[grf_.Prediction] predictions

    trainer = new grf_.ForestTrainer(
        grf_.instrumental_trainer(
            reduced_form_weight, stabilize_splits,
        )
    )

    data = Data(train_matrix)
    deref(data.c_data).set_outcome_index(outcome_index)
    deref(data.c_data).set_treatment_index(treatment_index)
    deref(data.c_data).set_instrument_index(treatment_index)
    if use_sample_weights:
        deref(data.c_data).set_weight_index(sample_weight_index)

    options = new grf_.ForestOptions(
        num_trees,
        ci_group_size,
        sample_fraction,
        mtry,
        min_node_size,
        honesty,
        honesty_fraction,
        honesty_prune_leaves,
        alpha,
        imbalance_penalty,
        num_threads,
        seed,
        clusters,
        samples_per_cluster,
    )

    forest = new grf_.Forest(trainer.train(deref(data.c_data), deref(options)))

    if compute_oob_predictions:
        predictor = new grf_.ForestPredictor(grf_.instrumental_predictor(num_threads))

    return create_forest_object(forest, predictions)


cpdef causal_predict(
    GRFForest forest_wrapper,
    np.ndarray[double, ndim=2, mode="fortran"] train_matrix,
    size_t outcome_index,
    size_t treatment_index,
    np.ndarray[double, ndim=2, mode="fortran"] test_matrix,
    unsigned int num_threads,
    bool estimate_variance,
):
    cdef vector[grf_.Prediction] predictions

    predictor = new grf_.ForestPredictor(
        grf_.instrumental_predictor(
            num_threads,
        )
    )
    train_data = Data(train_matrix)
    deref(train_data.c_data).set_outcome_index(outcome_index)
    deref(train_data.c_data).set_treatment_index(treatment_index)
    deref(train_data.c_data).set_instrument_index(treatment_index)

    test_data = Data(test_matrix)

    predictions = predictor.predict(
        deref(forest_wrapper.forest),
        deref(train_data.c_data),
        deref(test_data.c_data),
        estimate_variance,
    )

    return create_prediction_object(predictions)


cpdef probability_train(
    np.ndarray[double, ndim=2, mode="fortran"] train_matrix,
    size_t outcome_index,
    size_t sample_weight_index,
    bool use_sample_weights,
    size_t num_classes,
    unsigned int mtry,
    unsigned int num_trees,
    unsigned int min_node_size,
    double sample_fraction,
    bool honesty,
    double honesty_fraction,
    bool honesty_prune_leaves,
    size_t ci_group_size,
    double alpha,
    double imbalance_penalty,
    vector[size_t] clusters,
    unsigned int samples_per_cluster,
    bool compute_oob_predictions,
    unsigned int num_threads,
    unsigned int seed,
):
    cdef grf_.ForestOptions* options
    cdef vector[grf_.Prediction] predictions

    trainer = new grf_.ForestTrainer(grf_.probability_trainer(num_classes))
    data = Data(train_matrix)
    deref(data.c_data).set_outcome_index(outcome_index)

    if use_sample_weights:
        deref(data.c_data).set_weight_index(sample_weight_index - 1)

    options = new grf_.ForestOptions(
        num_trees,
        ci_group_size,
        sample_fraction,
        mtry,
        min_node_size,
        honesty,
        honesty_fraction,
        honesty_prune_leaves,
        alpha,
        imbalance_penalty,
        num_threads,
        seed,
        clusters,
        samples_per_cluster,
    )

    forest = new grf_.Forest(trainer.train(deref(data.c_data), deref(options)))

    if compute_oob_predictions:
        predictor = new grf_.ForestPredictor(grf_.probability_predictor(num_threads, num_classes))
        predictions = predictor.predict_oob(deref(forest), deref(data.c_data), False)

    return create_forest_object(forest, predictions)


cpdef probability_predict(
    GRFForest forest_wrapper,
    np.ndarray[double, ndim=2, mode="fortran"] train_matrix,
    size_t outcome_index,
    size_t num_classes,
    np.ndarray[double, ndim=2, mode="fortran"] test_matrix,
    unsigned int num_threads,
    bool estimate_variance,
):
    cdef vector[grf_.Prediction] predictions

    predictor = new grf_.ForestPredictor(grf_.probability_predictor(num_threads, num_classes))
    train_data = Data(train_matrix)
    deref(train_data.c_data).set_outcome_index(outcome_index)

    test_data = Data(test_matrix)

    predictions = predictor.predict(
        deref(forest_wrapper.forest),
        deref(train_data.c_data),
        deref(test_data.c_data),
        estimate_variance,
    )

    return create_prediction_object(predictions)



cdef create_forest_object(
    grf_.Forest* forest,
    vector[grf_.Prediction] predictions
):
    result = serialize_forest(forest)
    if not predictions.empty():
        result.update(create_prediction_object(predictions))
    return result


cdef grf_.Forest* deserialize_forest(
    dict forest_object
):
    cdef grf_.Forest* forest
    cdef vector[unique_ptr[grf_.Tree]] trees
    cdef grf_.Tree* tree
    cdef unique_ptr[grf_.Tree] treeptr

    cdef size_t num_trees = forest_object["num_trees"]

    trees.reserve(num_trees)

    for t in range(num_trees):
        tree = new grf_.Tree(
            forest_object["root_nodes"][t],
            forest_object["child_nodes"][t],
            forest_object["leaf_samples"][t],
            forest_object["split_vars"][t],
            forest_object["split_values"][t],
            forest_object["drawn_samples"][t],
            forest_object["send_missing_left"][t],
            grf_.PredictionValues(
                forest_object["pv_values"][t],
                forest_object["pv_num_types"][t]
            ),
        )
        treeptr.reset(tree)
        trees.push_back(
            move(treeptr)
        )
    forest = new grf_.Forest(
        trees,
        forest_object["num_variables"],
        forest_object["ci_group_size"]
    )
    return forest


cdef serialize_forest(
    grf_.Forest* forest,
):
    result = dict()
    result["ci_group_size"] = forest.get_ci_group_size()
    result["num_variables"] = forest.get_num_variables()
    num_trees = forest.get_trees().size()
    result["num_trees"] = num_trees

    result["root_nodes"] = []
    result["child_nodes"] = []
    result["leaf_samples"] = []
    result["split_vars"] = []
    result["split_values"] = []
    result["drawn_samples"] = []
    result["send_missing_left"] = []
    result["pv_values"] = []
    result["pv_num_types"] = []

    for k in range(num_trees):
        # this is awkward but it avoids a compilation issue on linux
        # related to overloading of get_trees_ :
        # error: call of overloaded ‘move(__gnu_cxx::__alloc_traits<std::allocator<std::unique_ptr<grf::Tree> > >::value_type&)’ is ambiguous
        #
        # assigning the tree doesn't work because Tree lacks a nullary constructor
        result["root_nodes"].append(deref(forest.get_trees()[k]).get_root_node())
        result["child_nodes"].append(deref(forest.get_trees()[k]).get_child_nodes())
        result["leaf_samples"].append(deref(forest.get_trees()[k]).get_leaf_samples())
        result["split_vars"].append(deref(forest.get_trees()[k]).get_split_vars())
        result["split_values"].append(deref(forest.get_trees()[k]).get_split_values())
        result["drawn_samples"].append(deref(forest.get_trees()[k]).get_drawn_samples())
        result["send_missing_left"].append(deref(forest.get_trees()[k]).get_send_missing_left())

        result["pv_values"].append(deref(forest.get_trees()[k]).get_prediction_values().get_all_values())
        result["pv_num_types"].append(deref(forest.get_trees()[k]).get_prediction_values().get_num_types())
    return result


cdef create_prediction_object(
    const vector[grf_.Prediction]& predictions,
):
    output = dict()
    output["predictions"] = create_prediction_matrix(predictions)
    output["variance_estimates"] = create_variance_matrix(predictions)
    output["debiased_error"] = create_error_matrix(predictions)
    output["excess_error"] = create_excess_error_matrix(predictions)
    return output


cdef create_prediction_matrix(
    const vector[grf_.Prediction]& predictions,
):
    cdef size_t prediction_length

    if predictions.empty():
        return []

    result = []
    for j in range(predictions.size()):
        values = []
        prediction_values = predictions[j].get_predictions()
        for k in range(prediction_values.size()):
            values.append(prediction_values[k])
        result.append(values)

    return result


cdef create_variance_matrix(
    const vector[grf_.Prediction]& predictions,
):
    cdef size_t prediction_length

    if predictions.empty():
        return []

    if not predictions.at(0).contains_variance_estimates():
        return []

    prediction_length = predictions.at(0).size()

    result = []
    for j in range(predictions.size()):
        values = []
        variance_estimate = predictions[j].get_variance_estimates()
        for k in range(variance_estimate.size()):
            values.append(variance_estimate[k])
        result.append(values)

    return result


cdef create_error_matrix(
    const vector[grf_.Prediction]& predictions,
):
    cdef size_t prediction_length

    if predictions.empty():
        return []

    if not predictions.at(0).contains_error_estimates():
        return []

    prediction_length = predictions.at(0).size()

    result = []
    for j in range(predictions.size()):
        values = []
        error_estimate = predictions[j].get_error_estimates()
        for k in range(error_estimate.size()):
            values.append(error_estimate[k])
        result.append(values)

    return result

cdef create_excess_error_matrix(
    const vector[grf_.Prediction]& predictions,
):
    cdef size_t prediction_length

    if predictions.empty():
        return []

    if not predictions.at(0).contains_error_estimates():
        return []

    prediction_length = predictions.at(0).size()

    result = []
    for j in range(predictions.size()):
        values = []
        error_estimate = predictions[j].get_excess_error_estimates()
        for k in range(error_estimate.size()):
            values.append(error_estimate[k])
        result.append(values)

    return result


cpdef compute_split_frequencies(
    GRFForest forest_wrapper,
    size_t max_depth,
):
    cdef grf_.SplitFrequencyComputer computer
    return computer.compute(deref(forest_wrapper.forest), max_depth)


cpdef compute_kernel_weights(
    GRFForest forest_wrapper,
    np.ndarray[double, ndim=2, mode="fortran"] test_matrix,
    int num_threads,
    bool oob_prediction,
):
    cdef grf_.SampleWeightComputer computer
    cdef unordered_map[size_t, double] calculated_weights
    cdef unordered_map[size_t, double].iterator it

    test_data = Data(test_matrix)

    tree_traverser = new grf_.TreeTraverser(num_threads)
    leaf_nodes_by_tree = tree_traverser.get_leaf_nodes(deref(forest_wrapper.forest), deref(test_data.c_data), oob_prediction)
    trees_by_sample = tree_traverser.get_valid_trees_by_sample(deref(forest_wrapper.forest), deref(test_data.c_data), oob_prediction)

    samples = []
    neighbors = []
    weights = []
    for sample in range(test_matrix.shape[0]):
        calculated_weights = computer.compute_weights(sample, deref(forest_wrapper.forest), leaf_nodes_by_tree, trees_by_sample)
        it = calculated_weights.begin()
        while (it != calculated_weights.end()):
            samples.append(sample)
            neighbors.append(deref(it).first)
            weights.append(deref(it).second)
            postinc(it)

    return samples, neighbors, weights





import logging
import numpy as np
from abc import ABC
from abc import abstractmethod
from numpy import random
from scipy import stats as ss
from sklearn.base import BaseEstimator
from sklearn.base import RegressorMixin
from sklearn.base import clone
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.utils.validation import check_array
from sklearn.utils.validation import check_is_fitted

from skgrf import grf
from skgrf.base import GRFMixin
from skgrf.ensemble.regressor import GRFForestRegressor
from skgrf.utils.validation import check_sample_weight

logger = logging.getLogger(__name__)


class GRFBoostedForestRegressor(GRFMixin, RegressorMixin, BaseEstimator):
    r"""GRF Boosted Regression implementation for sci-kit learn.

    Provides a sklearn regressor interface to the GRF C++ library using Cython.

    :param int n_estimators: The number of tree regressors to train
    :param bool equalize_cluster_weights: Weight the samples such that clusters have
        equally weight. If ``False``, larger clusters will have more weight. If
        ``True``, the number of samples drawn from each cluster is equal to the size of
        the smallest cluster. If ``True``, sample weights should not be passed on
        fitting.
    :param float sample_fraction: Fraction of samples used in each tree. If
        ``ci_group_size`` > 1, the max allowed fraction is 0.5
    :param int mtry: The number of features to split on each node. The default is
        ``sqrt(p) + 20`` where ``p`` is the number of features.
    :param int min_node_size: The minimum number of observations in each tree leaf.
    :param bool honesty: Use honest splitting (subsample splitting).
    :param float honesty_fraction: The fraction of data used for subsample splitting.
    :param bool honesty_prune_leaves: Prune estimation sample tree such that no leaves
        are empty. If ``False``, trees with empty leaves are skipped.
    :param float alpha: The maximum imbalance of a split.
    :param float imbalance_penalty: Penalty applied to imbalanced splits.
    :param int ci_group_size: The quantity of trees grown on each subsample. At least 2
        is required to provide confidence intervals.
    :param list(str) tune_params: A list of parameter names on which to perform tuning.
        Valid strings are "sample_fraction", "mtry", "min_node_size",
        "honesty_fraction", "honesty_prune_leaves", "alpha", "imbalance_penalty".
    :param int tune_n_estimators: The number of estimators to use in the tuning model.
    :param int tune_n_reps: The number of forests used in the tuning model
    :param int tune_n_draws: The number of random parameter values for tuning model
        selection
    :param int boost_steps: The number of boosting iterations
    :param int boost_error_reduction: The percentage of previous step's error that
        must be estimated by the current boost step
    :param int boost_max_steps: The maximum number of boosting iterations
    :param int boost_trees_tune: The number of trees used to test a new boosting step.
    :param int n_jobs: The number of threads. Default is number of CPU cores.
    :param int seed: Random seed value.

    :ivar int n_features_in\_: The number of features (columns) from the fit input
        ``X``.
    :ivar dict boosted_forests\_: The boosted regression forests.
    :ivar int mtry\_: The ``mtry`` value determined by validation.
    :ivar int outcome_index\_: The index of the grf train matrix holding the outcomes.
    :ivar list samples_per_cluster\_: The number of samples to train per cluster.
    :ivar list clusters\_: The cluster labels determined from the fit input ``cluster``.
    :ivar int n_clusters\_: The number of unique cluster labels from the fit input
        ``cluster``.
    """

    def __init__(
        self,
        n_estimators=100,
        equalize_cluster_weights=False,
        sample_fraction=0.5,
        mtry=None,
        min_node_size=5,
        honesty=True,
        honesty_fraction=0.5,
        honesty_prune_leaves=True,
        alpha=0.05,
        imbalance_penalty=0,
        ci_group_size=2,
        tune_params=None,
        tune_n_estimators=50,
        tune_n_reps=100,
        tune_n_draws=1000,
        boost_steps=None,
        boost_error_reduction=0.97,
        boost_max_steps=5,
        boost_trees_tune=10,
        n_jobs=-1,
        seed=42,
    ):
        self.n_estimators = n_estimators
        self.equalize_cluster_weights = equalize_cluster_weights
        self.sample_fraction = sample_fraction
        self.mtry = mtry
        self.min_node_size = min_node_size
        self.honesty = honesty
        self.honesty_fraction = honesty_fraction
        self.honesty_prune_leaves = honesty_prune_leaves
        self.alpha = alpha
        self.imbalance_penalty = imbalance_penalty
        self.ci_group_size = ci_group_size
        self.tune_params = tune_params
        self.tune_n_estimators = tune_n_estimators
        self.tune_n_reps = tune_n_reps
        self.tune_n_draws = tune_n_draws
        self.boost_steps = boost_steps
        self.boost_error_reduction = boost_error_reduction
        self.boost_max_steps = boost_max_steps
        self.boost_trees_tune = boost_trees_tune
        self.n_jobs = n_jobs
        self.seed = seed

    def fit(self, X, y, sample_weight=None, cluster=None):
        """Fit the grf forest using training data.

        :param array2d X: training input features
        :param array1d y: training input targets
        :param array1d sample_weight: optional weights for input samples
        :param array1d cluster: optional cluster assignments for input samples
        """
        X, y = self._validate_data(X, y)
        self._check_n_features(X, reset=True)

        self._check_boost_error_reduction()

        self._check_sample_fraction(oob=True)
        self._check_alpha()

        sample_weight, _ = check_sample_weight(sample_weight, X)

        cluster_ = self._check_cluster(X=X, cluster=cluster)
        self.samples_per_cluster_ = self._check_equalize_cluster_weights(
            cluster=cluster_, sample_weight=sample_weight
        )
        self.mtry_ = self._check_mtry(X=X)

        _ = self._create_train_matrices(X=X, y=y, sample_weight=sample_weight)

        # region tuning a regression forest
        regression_forest = GRFForestRegressor(
            n_estimators=self.tune_n_estimators,
            equalize_cluster_weights=self.equalize_cluster_weights,
            sample_fraction=self.sample_fraction,
            mtry=self.mtry,
            min_node_size=self.min_node_size,
            honesty=self.honesty,
            honesty_fraction=self.honesty_fraction,
            honesty_prune_leaves=self.honesty_prune_leaves,
            alpha=self.alpha,
            imbalance_penalty=self.imbalance_penalty,
            ci_group_size=self.ci_group_size,
            n_jobs=self.n_jobs,
            seed=self.seed,
        )
        if self.tune_params is None:
            logger.debug("not tuning boosted forest")
            regression_forest.fit(
                X=X,
                y=y,
                sample_weight=sample_weight,
                cluster=cluster,
                compute_oob_predictions=True,
            )
            params = regression_forest.get_params(deep=True)
            forest = regression_forest
        else:
            logger.debug("tuning boosted forest")
            tunable_params = (
                "sample_fraction",
                "mtry",
                "min_node_size",
                "honesty_fraction",
                "honesty_prune_leaves",
                "alpha",
                "imbalance_penalty",
            )
            param_distributions = {}
            for param in self.tune_params:
                if param not in tunable_params:
                    raise ValueError(
                        f"tuning param {param} not found in {str(tunable_params)}"
                    )
                param_distributions[param] = PARAM_DISTRIBUTIONS[param](*X.shape)

            uniform_samples = random.uniform(
                size=(self.tune_n_draws, len(self.tune_params))
            )
            param_samples = np.zeros(shape=(self.tune_n_draws, len(self.tune_params)))
            for idx, param in enumerate(self.tune_params):
                param_samples[:, idx] = param_distributions[param].dist(
                    uniform_samples[:, idx]
                )

            errors = []
            for draw in range(self.tune_n_draws):
                params = {
                    p: param_samples[draw, idx]
                    for idx, p in enumerate(self.tune_params)
                }
                regression_forest.set_params(**params)
                regression_forest.fit(
                    X=X,
                    y=y,
                    sample_weight=sample_weight,
                    cluster=cluster,
                    compute_oob_predictions=True,
                )
                errors.append(
                    np.nanmean(regression_forest.grf_forest_["debiased_error"])
                )

            if np.any(np.isnan(errors)):
                raise ValueError(
                    "unable to tune because of NaN-valued forest error estimates; consider more trees"
                )

            if np.std(errors) == 0 or np.std(errors) / np.mean(errors) < 1e-10:
                raise ValueError(
                    "unable to tune because of constant errors for forests; consider more trees"
                )

            variance_guess = np.var(errors) / 2
            gp = GaussianProcessRegressor(alpha=variance_guess)
            gp.fit(uniform_samples, errors)

            opt_samples = random.uniform(
                size=(self.tune_n_draws, len(self.tune_params))
            )

            model_surface = gp.predict(opt_samples)
            tuned_params = np.zeros(shape=(self.tune_n_draws, len(self.tune_params)))
            for idx, param in enumerate(self.tune_params):
                tuned_params[:, idx] = param_distributions[param].dist(
                    opt_samples[:, idx]
                )

            opt_idx = np.argmin(model_surface)
            params = {
                p: tuned_params[opt_idx, idx] for idx, p in enumerate(self.tune_params)
            }
            params.update(**{"n_estimators": self.tune_n_estimators * 4})
            regression_forest.set_params(**params)
            regression_forest.fit(
                X,
                y,
                sample_weight=sample_weight,
                cluster=cluster,
                compute_oob_predictions=True,
            )
            retrained_error = np.nanmean(
                regression_forest.grf_forest_["debiased_error"]
            )

            default_params = {
                "sample_fraction": 0.5,
                "mtry": min(np.ceil(np.sqrt(X.shape[1]) + 20), X.shape[1]),
                "min_node_size": 5,
                "honesty_fraction": 0.5,
                "honesty_prune_leaves": True,
                "alpha": 0.05,
                "imbalance_penalty": 0,
            }
            default_forest = clone(regression_forest)
            default_forest.set_params(**default_params)
            default_forest.fit(
                X=X,
                y=y,
                sample_weight=sample_weight,
                cluster=cluster,
                compute_oob_predictions=True,
            )
            default_error = np.nanmean(default_forest.grf_forest_["debiased_error"])

            if default_error < retrained_error:
                params = default_forest.get_params()
                forest = default_forest
            else:
                params = regression_forest.get_params()
                forest = regression_forest
        # endregion

        # region boosting with the tuned forest
        logger.debug("boosting forest")
        current_pred = {
            "predictions": forest.grf_forest_["predictions"],
            "debiased_error": forest.grf_forest_["debiased_error"],
            "excess_error": forest.grf_forest_["excess_error"],
        }

        y_hat = np.atleast_1d(np.squeeze(np.array(current_pred["predictions"])))
        debiased_error = current_pred["debiased_error"]
        boosted_forests = {
            "forest": [forest],
            "error": [np.mean(debiased_error)],
        }

        step = 1
        while True:
            y_residual = y - y_hat
            if self.boost_steps is not None:
                if step > self.boost_steps:
                    break
            elif step > self.boost_max_steps:
                break
            else:
                forest_small = GRFForestRegressor(
                    sample_fraction=params["sample_fraction"],
                    mtry=params["mtry"],
                    n_estimators=self.boost_trees_tune,
                    n_jobs=self.n_jobs,
                    min_node_size=params["min_node_size"],
                    honesty=self.honesty,
                    honesty_fraction=params["honesty_fraction"],
                    honesty_prune_leaves=params["honesty_prune_leaves"],
                    seed=self.seed,
                    ci_group_size=self.ci_group_size,
                    alpha=params["alpha"],
                    imbalance_penalty=params["imbalance_penalty"],
                    equalize_cluster_weights=self.equalize_cluster_weights,
                )
                forest_small.fit(
                    X=X,
                    y=y_residual,
                    sample_weight=sample_weight,
                    cluster=cluster,
                    compute_oob_predictions=True,
                )
                step_error = forest_small.grf_forest_["debiased_error"]
                if not np.nanmean(
                    step_error
                ) <= self.boost_error_reduction * np.nanmean(debiased_error):
                    break

            forest_residual = GRFForestRegressor(
                sample_fraction=params["sample_fraction"],
                mtry=params["mtry"],
                n_estimators=self.n_estimators,
                n_jobs=self.n_jobs,
                min_node_size=params["min_node_size"],
                honesty=self.honesty,
                honesty_fraction=params["honesty_fraction"],
                honesty_prune_leaves=params["honesty_prune_leaves"],
                seed=self.seed,
                ci_group_size=self.ci_group_size,
                alpha=params["alpha"],
                imbalance_penalty=params["imbalance_penalty"],
                equalize_cluster_weights=self.equalize_cluster_weights,
            )
            forest_residual.fit(
                X,
                y_residual,
                sample_weight=sample_weight,
                cluster=cluster,
                compute_oob_predictions=True,
            )
            current_pred = {
                "predictions": forest_residual.grf_forest_["predictions"],
                "debiased_error": forest_residual.grf_forest_["debiased_error"],
                "excess_error": forest_residual.grf_forest_["excess_error"],
            }
            y_hat = y_hat + np.atleast_1d(
                np.squeeze(np.array(current_pred["predictions"]))
            )
            debiased_error = current_pred["debiased_error"]
            boosted_forests["forest"].append(forest_residual)
            boosted_forests["error"].append(np.mean(debiased_error))
            step += 1
        # endregion

        boosted_forests["predictions"] = y_hat
        self.boosted_forests_ = boosted_forests
        return self

    def predict(self, X, boost_predict_steps=None):
        """Predict regression target for X.

        :param array2d X: prediction input features
        :param int boost_predict_steps: number of boost prediction steps
        """
        check_is_fitted(self)
        X = check_array(X)
        self._check_n_features(X, reset=False)
        num_forests = len(self.boosted_forests_["forest"])
        if boost_predict_steps is None:
            boost_predict_steps = num_forests
        else:
            boost_predict_steps = min(boost_predict_steps, num_forests)

        y_hat = 0
        for forest in self.boosted_forests_["forest"][:boost_predict_steps]:
            forest._ensure_ptr()
            result = grf.regression_predict(
                forest.grf_forest_cpp_,
                np.asfortranarray([[]]),  # train_matrix
                self.outcome_index_,
                np.asfortranarray(X.astype("float64")),  # test_matrix
                self._get_num_threads(),
                False,  # estimate_variance
            )
            y = np.atleast_1d(np.squeeze(np.array(result["predictions"])))
            y_hat = y_hat + y
        return y_hat

    def _more_tags(self):
        return {
            "requires_y": True,
            "_xfail_checks": {
                "check_sample_weights_invariance": "zero sample_weight is not equivalent to removing samples",
            },
        }


# region random parameter tuning functions
class GRFParamDistribution(ABC):
    def __init__(self, X_rows: int, X_cols: int):
        self.X_rows = X_rows
        self.X_cols = X_cols

    @abstractmethod
    def rvs(self, *args, **kwds):  # pragma: no cover
        raise NotImplementedError()


class GRFMinNodeSizeDistribution(GRFParamDistribution):
    def rvs(self, *args, **kwds):  # pragma: no cover
        return self.dist(ss.uniform(*args, **kwds))

    def dist(self, uniform):
        return np.floor(np.exp2(uniform * (np.log(self.X_rows) - np.log(2) - 4)))


class GRFSampleFractionDistribution(GRFParamDistribution):
    def rvs(self, *args, **kwds):  # pragma: no cover
        return self.dist(ss.uniform(*args, **kwds))

    def dist(self, uniform):
        return 0.05 + 0.45 * uniform


class GRFMtryDistribution(GRFParamDistribution):
    def rvs(self, *args, **kwds):  # pragma: no cover
        return self.dist(ss.uniform(*args, **kwds))

    def dist(self, uniform):
        return np.ceil(np.min([self.X_cols, np.sqrt(self.X_cols) + 20]) * uniform)


class GRFAlphaDistribution(GRFParamDistribution):
    def rvs(self, *args, **kwds):  # pragma: no cover
        return self.dist(ss.uniform(*args, **kwds))

    def dist(self, uniform):
        return uniform / 4


class GRFImbalancePenaltyDistribution(GRFParamDistribution):
    def rvs(self, *args, **kwds):  # pragma: no cover
        return self.dist(ss.uniform(*args, **kwds))

    def dist(self, uniform):
        return -np.log(uniform)


class GRFHonestyFractionDistribution(GRFParamDistribution):
    def rvs(self, *args, **kwds):  # pragma: no cover
        return self.dist(ss.uniform(*args, **kwds))

    def dist(self, uniform):
        return 0.5 + 0.3 * uniform


class GRFHonestyPruneLeavesDistribution(GRFParamDistribution):
    def __init__(self, X_rows: int, X_cols: int):
        super().__init__(X_rows, X_cols)
        self.choices = np.array([True, False])

    def rvs(self, *args, **kwds):  # pragma: no cover
        return self.dist(ss.uniform(*args, **kwds))

    def dist(self, uniform):
        return self.choices[(uniform < 0.5).astype(int)]


PARAM_DISTRIBUTIONS = {
    "min_node_size": GRFMinNodeSizeDistribution,
    "sample_fraction": GRFSampleFractionDistribution,
    "mtry": GRFMtryDistribution,
    "alpha": GRFAlphaDistribution,
    "imbalance_penalty": GRFImbalancePenaltyDistribution,
    "honesty_fraction": GRFHonestyFractionDistribution,
    "honesty_prune_leaves": GRFHonestyPruneLeavesDistribution,
}
# endregion



import logging
from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_is_fitted

from skgrf.ensemble.boosted_regressor import GRFBoostedForestRegressor
from skgrf.ensemble.instrumental_regressor import GRFForestInstrumentalRegressor
from skgrf.tree.causal_regressor import GRFTreeCausalRegressor

logger = logging.getLogger(__name__)


class GRFForestCausalRegressor(GRFForestInstrumentalRegressor):
    r"""GRF Causal regression implementation for sci-kit learn.

    Provides a sklearn causal regressor to the GRF C++ library using Cython.

    :param int n_estimators: The number of tree regressors to train
    :param bool equalize_cluster_weights: Weight the samples such that clusters have
        equally weight. If ``False``, larger clusters will have more weight. If
        ``True``, the number of samples drawn from each cluster is equal to the size of
        the smallest cluster. If ``True``, sample weights should not be passed on
        fitting.
    :param float sample_fraction: Fraction of samples used in each tree. If
        ``ci_group_size`` > 1, the max allowed fraction is 0.5
    :param int mtry: The number of features to split on each node. The default is
        ``sqrt(p) + 20`` where ``p`` is the number of features.
    :param int min_node_size: The minimum number of observations in each tree leaf.
    :param bool honesty: Use honest splitting (subsample splitting).
    :param float honesty_fraction: The fraction of data used for subsample splitting.
    :param bool honesty_prune_leaves: Prune estimation sample tree such that no leaves
        are empty. If ``False``, trees with empty leaves are skipped.
    :param float alpha: The maximum imbalance of a split.
    :param float imbalance_penalty: Penalty applied to imbalanced splits.
    :param int ci_group_size: The quantity of trees grown on each subsample. At least 2
        is required to provide confidence intervals.
    :param bool orthogonal_boosting: When ``y_hat`` or ``w_hat`` are ``None``, they
        are estimated using boosted regression forests. (Not yet implemented)
    :param bool stabilize_splits: Whether or not the instrument should be taken into
        account when determining the imbalance of a split.
    :param int n_jobs: The number of threads. Default is number of CPU cores.
    :param int seed: Random seed value.
    :param bool enable_tree_details: When ``True``, perform additional calculations
        for detailing the underlying decision trees. Must be enabled for ``estimators_``
        and ``get_estimator`` to work. Very slow.

    :ivar list estimators\_: A list of tree objects from the forest.
    :ivar int n_features_in\_: The number of features (columns) from the fit input
        ``X``.
    :ivar dict grf_forest\_: The returned result object from calling C++ grf.
    :ivar int mtry\_: The ``mtry`` value determined by validation.
    :ivar int outcome_index\_: The index of the grf train matrix holding the outcomes.
    :ivar list samples_per_cluster\_: The number of samples to train per cluster.
    :ivar list clusters\_: The cluster labels determined from the fit input ``cluster``.
    :ivar int n_clusters\_: The number of unique cluster labels from the fit input
        ``cluster``.
    :ivar str criterion: The criterion used for splitting: ``mse``
    """

    def __init__(
        self,
        n_estimators=2000,
        equalize_cluster_weights=False,
        sample_fraction=0.5,
        mtry=None,
        min_node_size=5,
        honesty=True,
        honesty_fraction=0.5,
        honesty_prune_leaves=True,
        alpha=0.05,
        imbalance_penalty=0,
        ci_group_size=2,
        stabilize_splits=True,
        orthogonal_boosting=False,
        n_jobs=-1,
        seed=42,
        enable_tree_details=False,
    ):
        super().__init__(
            n_estimators=n_estimators,
            equalize_cluster_weights=equalize_cluster_weights,
            sample_fraction=sample_fraction,
            mtry=mtry,
            min_node_size=min_node_size,
            honesty=honesty,
            honesty_fraction=honesty_fraction,
            honesty_prune_leaves=honesty_prune_leaves,
            alpha=alpha,
            imbalance_penalty=imbalance_penalty,
            ci_group_size=ci_group_size,
            reduced_form_weight=0,
            stabilize_splits=stabilize_splits,
            n_jobs=n_jobs,
            seed=seed,
            enable_tree_details=enable_tree_details,
        )
        self.orthogonal_boosting = orthogonal_boosting

    @property
    def estimators_(self):
        try:
            check_is_fitted(self)
        except NotFittedError:
            raise AttributeError(
                f"{self.__class__.__name__} object has no attribute 'estimators_'"
            ) from None
        if not self.enable_tree_details:
            raise ValueError("enable_tree_details must be True prior to training")
        return [
            GRFTreeCausalRegressor.from_forest(self, idx=idx)
            for idx in range(self.n_estimators)
        ]

    def get_estimator(self, idx):
        """Extract a single estimator tree from the forest.

        :param int idx: The index of the tree to extract.
        """
        check_is_fitted(self)
        if not self.enable_tree_details:
            raise ValueError("enable_tree_details must be True prior to training")
        return GRFTreeCausalRegressor.from_forest(self, idx=idx)

    # noinspection PyMethodOverriding
    def fit(
        self,
        X,
        y,
        w,  # treatment
        y_hat=None,
        w_hat=None,
        sample_weight=None,
        cluster=None,
    ):
        """Fit the grf forest using training data.

        :param array2d X: training input features
        :param array1d y: training input targets
        :param array1d w: training input treatments
        :param array1d y_hat: estimated expected target responses
        :param array1d w_hat: estimated treatment propensities
        :param array1d sample_weight: optional weights for input samples
        :param array1d cluster: optional cluster assignments for input samples
        """
        X, y = self._validate_data(X, y)
        self._check_num_samples(X)

        boost_params = {
            "n_estimators": max(50, int(self.n_estimators / 4)),
            "equalize_cluster_weights": self.equalize_cluster_weights,
            "sample_fraction": self.sample_fraction,
            "mtry": self.mtry,
            "min_node_size": 5,
            "honesty": True,
            "honesty_fraction": 0.5,
            "honesty_prune_leaves": self.honesty_prune_leaves,
            "alpha": self.alpha,
            "imbalance_penalty": self.imbalance_penalty,
            "ci_group_size": 1,
            "tune_params": None,  # TODO ?
            "n_jobs": self.n_jobs,
            "seed": self.seed,
        }
        if y_hat is None and self.orthogonal_boosting:
            logger.debug("orthogonal boosting y_hat")
            br = GRFBoostedForestRegressor(**boost_params)
            br.fit(X, y, sample_weight=sample_weight, cluster=cluster)
            y_hat = br.boosted_forests_["predictions"]

        if w_hat is None and self.orthogonal_boosting:
            logger.debug("orthogonal boosting w_hat")
            br = GRFBoostedForestRegressor(**boost_params)
            br.fit(X, w, sample_weight=sample_weight, cluster=cluster)
            w_hat = br.boosted_forests_["predictions"]

        return super().fit(
            X=X,
            y=y,
            w=w,
            z=w,
            y_hat=y_hat,
            w_hat=w_hat,
            z_hat=w_hat,
            cluster=cluster,
            sample_weight=sample_weight,
        )



import numpy as np
from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_is_fitted

from skgrf import grf


class GRFMixin:
    def __getstate__(self):
        """Serialize.

        The GRFForest pointer ref is not serializable, so
        we must pop it off of state here.
        """
        self.__dict__.pop("grf_forest_cpp_", None)
        return self.__dict__

    def __setstate__(self, state):
        """Deserialize.

        Unpickle and ensure that we deserialize the forest
        to a C++ pointer, so that predictions are fast.
        """
        self.__dict__ = state
        try:
            check_is_fitted(self)
            self._ensure_ptr()
        except NotFittedError:
            pass

    def _ensure_ptr(self):
        """Ensure that a pointer to the C++ forest exists."""
        if hasattr(self, "grf_forest_") and not hasattr(self, "grf_forest_cpp_"):
            self.grf_forest_cpp_ = grf.GRFForest(self.grf_forest_)

    def _check_cluster(self, X, cluster):
        """Validate cluster definitions against training data."""
        if cluster is None:
            self.clusters_ = None
            self.n_clusters_ = 1
            return np.array([])
        if len(cluster) != X.shape[0]:
            raise ValueError("cluster length must be the same as X")
        self.clusters_, new_cluster = np.unique(cluster, return_inverse=True)
        self.n_clusters_ = len(self.clusters_)
        return new_cluster

    def _check_equalize_cluster_weights(self, cluster, sample_weight):
        """Validate cluster weight against equalize cluster weight param."""
        if len(cluster) == 0:
            return 0
        _, counts = np.unique(cluster, return_counts=True)
        if self.equalize_cluster_weights:
            if sample_weight is not None:
                raise ValueError(
                    "Cannot use sample_weight when equalize_cluster_weights is True"
                )
            return min(counts)
        else:
            return max(counts)

    def _check_mtry(self, X):
        """Validate mtry."""
        if self.mtry is None:
            return min(int(np.ceil(np.sqrt(X.shape[1] + 20))), X.shape[1])
        else:
            return self.mtry

    def _check_sample_fraction(self, oob=False):
        """Validate sample fraction"""
        if hasattr(self, "ci_group_size") and self.ci_group_size >= 2:
            if self.sample_fraction <= 0 or self.sample_fraction > 0.5:
                raise ValueError(
                    "sample fraction must be between 0 and 0.5 when ci_group_size >= 2"
                )
        if self.sample_fraction <= 0 or self.sample_fraction > 1:
            raise ValueError(
                "sample fraction must be between 0 and 1 when ci_group_size == 1"
            )
        if oob and self.sample_fraction >= 1:
            raise ValueError(
                "sample fraction must be strictly less than 1 for oob predictions"
            )

    def _check_alpha(self):
        """Validate alpha."""
        if self.alpha <= 0 or self.alpha >= 0.25:
            raise ValueError("alpha must be between 0 and 0.25")

    def _check_reduced_form_weight(self):
        """Validate reduced form weight."""
        if self.reduced_form_weight < 0 or self.reduced_form_weight > 1:
            raise ValueError("reduced_form_weight must be between 0 and 1")

    def _check_boost_error_reduction(self):
        """Validate boost error reduction."""
        if self.boost_error_reduction < 0 or self.boost_error_reduction > 1:
            raise ValueError("reduced_form_weight must be between 0 and 1")

    def _get_num_threads(self):
        """Get GRF-expected num_threads value."""
        return max([self.n_jobs, 0])  # sklearn convention is -1 for all cpus, grf is 0

    def _create_train_matrices(
        self,
        X,
        y=None,
        sample_weight=None,
        treatment=None,
        instrument=None,
        censor=None,
    ):
        """Create a concatenated training matrix.

        GRF expects training data to be combined into a single matrix with reference variables
        that point to column indices. This creates that concatenated matrix and sets
        reference indices to be passed to GRF.
        """
        n_cols = X.shape[1]
        concats = [X]
        if y is not None:
            self.outcome_index_ = n_cols
            n_cols += 1
            concats.append(np.atleast_2d(y).T)
        if treatment is not None:
            self.treatment_index_ = n_cols
            n_cols += 1
            concats.append(np.atleast_2d(treatment).T)
        if instrument is not None:
            self.instrument_index_ = n_cols
            n_cols += 1
            concats.append(np.atleast_2d(instrument).T)
        if censor is not None:
            self.censor_index_ = n_cols
            n_cols += 1
            concats.append(np.atleast_2d(censor).T)
        self.sample_weight_index_ = 0
        if sample_weight is not None:
            self.sample_weight_index_ = n_cols
            n_cols += 1
            concats.append(np.atleast_2d(sample_weight).T)
        return np.concatenate(concats, axis=1)

    def _check_num_samples(self, X):
        """Validate sample fraction against dataset."""
        if len(X) * self.sample_fraction < 1:
            raise ValueError(
                "The sample fraction is too small, resulting in less than 1 sample for fitting."
            )

    def _set_sample_weights(self, sample_weights):
        """Set leaf weights for access in ``Tree``."""
        weights = []
        for tree in sample_weights:
            tree_weights = []
            for idx, node in enumerate(tree):
                tree_weights.append(sum(node))
            weights.append(tree_weights)
        self.grf_forest_["leaf_weights"] = weights

    def _get_sample_values(self, values):
        """Map the leaf samples to corresponding values.

        Used to get corresponding sample weights or targets.
        """
        mapped_values = []
        for tree in self.grf_forest_["leaf_samples"]:
            mapped_values.append([])
            for node in tree:
                mapped_values[-1].append([values[idx] for idx in node])
        return mapped_values

    def _get_values(self, left, right, idx, values):
        """Recursively sum the child values through a tree."""
        left_values = (
            values[idx]
            if left[idx] == 0
            else self._get_values(left, right, left[idx], values)
        )
        right_values = (
            values[idx]
            if right[idx] == 0
            else self._get_values(left, right, right[idx], values)
        )
        values[idx] = left_values + right_values
        return values[idx]

    def _set_node_values(self, y, sample_weight):
        """Set the node values for ``Tree.value``."""
        forest_values = self._get_sample_values(y)
        forest_weights = self._get_sample_values(sample_weight)
        self._set_sample_weights(forest_weights)
        self.grf_forest_["node_values"] = []
        for idx in range(self.grf_forest_["num_trees"]):
            left = self.grf_forest_["child_nodes"][idx][0]
            right = self.grf_forest_["child_nodes"][idx][1]
            root = self.grf_forest_["root_nodes"][idx]
            values = forest_values[idx]
            self._get_values(
                left,
                right,
                root,
                values,
            )
            weights = forest_weights[idx]
            self._get_values(
                left,
                right,
                root,
                weights,
            )
            values = [
                np.average(v, weights=w, axis=0) if v else np.nan
                for v, w in zip(values, weights)
            ]
            self.grf_forest_["node_values"].append(values)

    def _set_n_classes(self):
        """Set num classes for ``Tree.n_classes``."""
        # for accessing in Tree
        self.grf_forest_["n_classes"] = getattr(self, "n_classes_", 1)



import numpy as np
from sklearn.base import BaseEstimator
from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_array
from sklearn.utils.validation import check_is_fitted

from skgrf import grf
from skgrf.ensemble.base import BaseGRFForest
from skgrf.tree.survival import GRFTreeSurvival
from skgrf.utils.validation import check_sample_weight


class GRFForestSurvival(BaseGRFForest, BaseEstimator):
    r"""GRF Survival implementation for sci-kit learn.

    Provides a sklearn survival interface to the GRF C++ library using Cython.

    .. warning::

        Because the training dataset is required for prediction, the training dataset
        is recorded onto the estimator instance. This means that serializing this
        estimator will result in a file at least as large as the serialized training
        dataset.

    :param int n_estimators: The number of survival trees to train
    :param bool equalize_cluster_weights: Weight the samples such that clusters have
        equally weight. If ``False``, larger clusters will have more weight. If
        ``True``, the number of samples drawn from each cluster is equal to the size of
        the smallest cluster. If ``True``, sample weights should not be passed on
        fitting.
    :param float sample_fraction: Fraction of samples used in each tree.
    :param int mtry: The number of features to split on each node. The default is
        ``sqrt(p) + 20`` where ``p`` is the number of features.
    :param int min_node_size: The minimum number of observations in each tree leaf.
    :param bool honesty: Use honest splitting (subsample splitting).
    :param float honesty_fraction: The fraction of data used for subsample splitting.
    :param bool honesty_prune_leaves: Prune estimation sample tree such that no leaves
        are empty. If ``False``, trees with empty leaves are skipped.
    :param float alpha: The maximum imbalance of a split.
    :param int n_jobs: The number of threads. Default is number of CPU cores.
    :param int seed: Random seed value.
    :param bool enable_tree_details: When ``True``, perform additional calculations
        for detailing the underlying decision trees. Must be enabled for ``estimators_``
        and ``get_estimator`` to work. Very slow.

    :ivar list estimators\_: A list of tree objects from the forest.
    :ivar int n_features_in\_: The number of features (columns) from the fit input
        ``X``.
    :ivar dict grf_forest\_: The returned result object from calling C++ grf.
    :ivar int mtry\_: The ``mtry`` value determined by validation.
    :ivar int outcome_index\_: The index of the grf train matrix holding the outcomes.
    :ivar int censor_index\_: The index of the grf train matrix holding the censoring.
    :ivar array1d failure_times_\_: An array of unique failure times from the training
        set.
    :ivar int num_failures_\_: The length of the ``failure_times`` array.
    :ivar list clusters\_: The cluster labels determined from the fit input ``cluster``.
    :ivar int n_clusters\_: The number of unique cluster labels from the fit input
        ``cluster``.
    :ivar str criterion: The criterion used for splitting: ``logrank``
    """

    def __init__(
        self,
        n_estimators=100,
        equalize_cluster_weights=False,
        sample_fraction=0.5,
        mtry=None,
        min_node_size=5,
        honesty=True,
        honesty_fraction=0.5,
        honesty_prune_leaves=True,
        alpha=0.05,
        n_jobs=-1,
        seed=42,
        enable_tree_details=False,
    ):
        self.n_estimators = n_estimators
        self.equalize_cluster_weights = equalize_cluster_weights
        self.sample_fraction = sample_fraction
        self.mtry = mtry
        self.min_node_size = min_node_size
        self.honesty = honesty
        self.honesty_fraction = honesty_fraction
        self.honesty_prune_leaves = honesty_prune_leaves
        self.alpha = alpha
        self.n_jobs = n_jobs
        self.seed = seed
        self.enable_tree_details = enable_tree_details

    @property
    def criterion(self):
        return "logrank"

    @property
    def estimators_(self):
        try:
            check_is_fitted(self)
        except NotFittedError:
            raise AttributeError(
                f"{self.__class__.__name__} object has no attribute 'estimators_'"
            ) from None
        if not self.enable_tree_details:
            raise ValueError("enable_tree_details must be True prior to training")
        return [
            GRFTreeSurvival.from_forest(self, idx=idx)
            for idx in range(self.n_estimators)
        ]

    def get_estimator(self, idx):
        """Extract a single estimator tree from the forest.

        :param int idx: The index of the tree to extract.
        """
        check_is_fitted(self)
        if not self.enable_tree_details:
            raise ValueError("enable_tree_details must be True prior to training")
        return GRFTreeSurvival.from_forest(self, idx=idx)

    def fit(self, X, y, sample_weight=None, cluster=None):
        """Fit the grf forest using training data.

        :param array2d X: training input features
        :param array1d y: training input targets, rows of (bool, float) representing
            (survival, time)
        :param array1d sample_weight: optional weights for input samples
        :param array1d cluster: optional cluster assignments for input samples
        """
        X = check_array(X)
        self._check_num_samples(X)
        self._check_n_features(X, reset=True)
        y = np.array(y.tolist())

        self._check_sample_fraction()
        self._check_alpha()

        cluster = self._check_cluster(X=X, cluster=cluster)
        self.samples_per_cluster_ = self._check_equalize_cluster_weights(
            cluster=cluster, sample_weight=sample_weight
        )

        sample_weight, use_sample_weight = check_sample_weight(sample_weight, X)

        self.mtry_ = self._check_mtry(X=X)

        # Extract the failure times from the training targets
        self.failure_times_ = np.sort(np.unique(y[:, 1][y[:, 0] == 1]))
        self.num_failures_ = len(self.failure_times_)

        # Relabel the failure times to consecutive integers
        y_times_relabeled = np.searchsorted(self.failure_times_, y[:, 1])
        y_censor = y[:, 0]

        train_matrix = self._create_train_matrices(
            X, y_times_relabeled, sample_weight=sample_weight, censor=y_censor
        )
        self.train_ = train_matrix

        self.grf_forest_ = grf.survival_train(
            np.asfortranarray(train_matrix.astype("float64")),
            self.outcome_index_,
            self.censor_index_,
            self.sample_weight_index_,
            use_sample_weight,
            self.mtry_,
            self.n_estimators,  # num_trees
            self.min_node_size,
            self.sample_fraction,
            self.honesty,
            self.honesty_fraction,
            self.honesty_prune_leaves,
            self.alpha,
            self.num_failures_,
            cluster,
            self.samples_per_cluster_,
            False,  # compute_oob_predictions,
            self._get_num_threads(),  # num_threads,
            self.seed,
        )
        self._ensure_ptr()

        if self.enable_tree_details:
            sample_weight = (
                sample_weight if sample_weight is not None else np.ones(len(X))
            )
            self._set_node_values(y, sample_weight)
            self._set_n_classes()

        return self

    def predict_cumulative_hazard_function(self, X):
        """Predict cumulative hazard function.

        :param array2d X: prediction input features
        """
        surv = self.predict_survival_function(X)
        return -np.log(surv)

    def predict(self, X):
        """Predict risk score.

        :param array2d X: prediction input features
        """
        chf = self.predict_cumulative_hazard_function(X)
        return chf.sum(1)

    def predict_survival_function(self, X):
        """Predict survival function.

        :param array2d X: prediction input features
        """
        return np.atleast_1d(np.squeeze(np.array(self._predict(X)["predictions"])))

    def _predict(self, X):
        check_is_fitted(self)
        X = check_array(X)
        self._check_n_features(X, reset=False)
        self._ensure_ptr()

        result = grf.survival_predict(
            self.grf_forest_cpp_,
            np.asfortranarray(self.train_.astype("float64")),  # test_matrix
            self.outcome_index_,
            self.censor_index_,
            self.sample_weight_index_,
            False,  # use_sample_weights
            np.asfortranarray(X.astype("float64")),  # test_matrix
            self._get_num_threads(),
            self.num_failures_,
        )
        return result

    def _more_tags(self):
        return {
            "requires_y": True,
            "_xfail_checks": {
                "check_sample_weights_invariance": "zero sample_weight is not equivalent to removing samples",
            },
        }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['skgrf', 'skgrf.ensemble', 'skgrf.tree', 'skgrf.utils']

package_data = \
{'': ['*'],
 'skgrf': ['grf/*',
           'grf/src/*',
           'grf/src/analysis/*',
           'grf/src/commons/*',
           'grf/src/forest/*',
           'grf/src/prediction/*',
           'grf/src/prediction/collector/*',
           'grf/src/relabeling/*',
           'grf/src/sampling/*',
           'grf/src/splitting/*',
           'grf/src/splitting/factory/*',
           'grf/src/tree/*',
           'grf/test/*',
           'grf/test/analysis/*',
           'grf/test/commons/*',
           'grf/test/forest/*',
           'grf/test/forest/resources/*',
           'grf/test/prediction/*',
           'grf/test/relabeling/*',
           'grf/test/sampling/*',
           'grf/test/splitting/*',
           'grf/test/splitting/resources/*',
           'grf/test/tree/*',
           'grf/test/utilities/*',
           'grf/third_party/Eigen/*',
           'grf/third_party/Eigen/src/Cholesky/*',
           'grf/third_party/Eigen/src/CholmodSupport/*',
           'grf/third_party/Eigen/src/Core/*',
           'grf/third_party/Eigen/src/Core/arch/AVX/*',
           'grf/third_party/Eigen/src/Core/arch/AVX512/*',
           'grf/third_party/Eigen/src/Core/arch/AltiVec/*',
           'grf/third_party/Eigen/src/Core/arch/CUDA/*',
           'grf/third_party/Eigen/src/Core/arch/Default/*',
           'grf/third_party/Eigen/src/Core/arch/NEON/*',
           'grf/third_party/Eigen/src/Core/arch/SSE/*',
           'grf/third_party/Eigen/src/Core/arch/ZVector/*',
           'grf/third_party/Eigen/src/Core/functors/*',
           'grf/third_party/Eigen/src/Core/products/*',
           'grf/third_party/Eigen/src/Core/util/*',
           'grf/third_party/Eigen/src/Eigenvalues/*',
           'grf/third_party/Eigen/src/Geometry/*',
           'grf/third_party/Eigen/src/Geometry/arch/*',
           'grf/third_party/Eigen/src/Householder/*',
           'grf/third_party/Eigen/src/IterativeLinearSolvers/*',
           'grf/third_party/Eigen/src/Jacobi/*',
           'grf/third_party/Eigen/src/LU/*',
           'grf/third_party/Eigen/src/LU/arch/*',
           'grf/third_party/Eigen/src/MetisSupport/*',
           'grf/third_party/Eigen/src/OrderingMethods/*',
           'grf/third_party/Eigen/src/PaStiXSupport/*',
           'grf/third_party/Eigen/src/PardisoSupport/*',
           'grf/third_party/Eigen/src/QR/*',
           'grf/third_party/Eigen/src/SPQRSupport/*',
           'grf/third_party/Eigen/src/SVD/*',
           'grf/third_party/Eigen/src/SparseCholesky/*',
           'grf/third_party/Eigen/src/SparseCore/*',
           'grf/third_party/Eigen/src/SparseLU/*',
           'grf/third_party/Eigen/src/SparseQR/*',
           'grf/third_party/Eigen/src/StlSupport/*',
           'grf/third_party/Eigen/src/SuperLUSupport/*',
           'grf/third_party/Eigen/src/UmfPackSupport/*',
           'grf/third_party/Eigen/src/misc/*',
           'grf/third_party/Eigen/src/plugins/*',
           'grf/third_party/optional/*',
           'grf/third_party/random/*']}

install_requires = \
['scikit-learn>=0.23.0,<1.0']

setup_kwargs = {
    'name': 'skgrf',
    'version': '0.2.0',
    'description': 'python bindings for C++ generalized random forests (grf)',
    'long_description': "skgrf\n=====\n\n|actions| |wheels| |rtd| |pypi| |pyversions|\n\n.. |actions| image:: https://github.com/crflynn/skgrf/workflows/build/badge.svg\n    :target: https://github.com/crflynn/skgrf/actions\n\n.. |wheels| image:: https://github.com/crflynn/skgrf-wheels/workflows/wheels/badge.svg\n    :target: https://github.com/crflynn/skgrf-wheels/actions\n\n.. |rtd| image:: https://img.shields.io/readthedocs/skgrf.svg\n    :target: http://skgrf.readthedocs.io/en/latest/\n\n.. |pypi| image:: https://img.shields.io/pypi/v/skgrf.svg\n    :target: https://pypi.python.org/pypi/skgrf\n\n.. |pyversions| image:: https://img.shields.io/pypi/pyversions/skgrf.svg\n    :target: https://pypi.python.org/pypi/skgrf\n\n``skgrf`` provides `scikit-learn <https://scikit-learn.org/stable/index.html>`__ compatible Python bindings to the C++ random forest implementation, `grf <https://github.com/grf-labs/grf>`__, using `Cython <https://cython.readthedocs.io/en/latest/>`__.\n\nThe latest release of ``skgrf`` uses version `2.0.0 <https://github.com/grf-labs/grf/releases/tag/v2.0.0>`__ of ``grf``.\n\n``skgrf`` is still in development. Please create issues for any discrepancies or errors. PRs welcome.\n\nDocumentation\n-------------\n\n* `stable <https://skgrf.readthedocs.io/en/stable/>`__\n* `latest <https://skgrf.readthedocs.io/en/latest/>`__\n\nInstallation\n------------\n\n``skgrf`` is available on `pypi <https://pypi.org/project/skgrf>`__ and can be installed via pip:\n\n.. code-block:: bash\n\n    pip install skgrf\n\nEstimators\n----------\n\n* GRFForestCausalRegressor\n* GRFForestInstrumentalRegressor\n* GRFForestLocalLinearRegressor\n* GRFForestQuantileRegressor\n* GRFForestRegressor\n* GRFBoostedForestRegressor\n* GRFForestSurvival\n\nUsage\n-----\n\nGRFForestRegressor\n~~~~~~~~~~~~~~~~~~\n\nThe ``GRFForestRegressor`` predictor uses ``grf``'s RegressionPredictionStrategy class.\n\n.. code-block:: python\n\n    from sklearn.datasets import load_boston\n    from sklearn.model_selection import train_test_split\n    from skgrf.ensemble import GRFForestRegressor\n    \n    X, y = load_boston(return_X_y=True)\n    X_train, X_test, y_train, y_test = train_test_split(X, y)\n    \n    forest = GRFForestRegressor()\n    forest.fit(X_train, y_train)\n    \n    predictions = forest.predict(X_test)\n    print(predictions)\n    # [31.81349144 32.2734354  16.51560285 11.90284392 39.69744341 21.30367911\n    #  19.52732937 15.82126562 26.49528961 11.27220097 16.02447197 20.01224404\n    #  ...\n    #  20.70674263 17.09041289 12.89671205 20.79787926 21.18317924 25.45553279\n    #  20.82455595]\n\n\nGRFForestQuantileRegressor\n~~~~~~~~~~~~~~~~~~~~~~~~~~\n\nThe ``GRFForestQuantileRegressor`` predictor uses ``grf``'s QuantilePredictionStrategy class.\n\n.. code-block:: python\n\n    from sklearn.datasets import load_boston\n    from sklearn.model_selection import train_test_split\n    from skgrf.ensemble import GRFForestQuantileRegressor\n    \n    X, y = load_boston(return_X_y=True)\n    X_train, X_test, y_train, y_test = train_test_split(X, y)\n    \n    forest = GRFForestQuantileRegressor(quantiles=[0.1, 0.9])\n    forest.fit(X_train, y_train)\n    \n    predictions = forest.predict(X_test)\n    print(predictions)\n    # [[21.9 50. ]\n    # [ 8.5 24.5]\n    # ...\n    # [ 8.4 18.6]\n    # [ 8.1 20. ]]\n\nLicense\n-------\n\n``skgrf`` is licensed under `GPLv3 <https://github.com/crflynn/skgrf/blob/main/LICENSE.txt>`__.\n\nDevelopment\n-----------\n\nTo develop locally, it is recommended to have ``asdf``, ``make`` and a C++ compiler already installed. After cloning, run ``make setup``. This will setup the grf submodule, install python and poetry from ``.tool-versions``, install dependencies using poetry, copy the grf source code into skgrf, and then build and install skgrf in the local virtualenv.\n\nTo format code, run ``make fmt``. This will run isort and black against the .py files.\n\nTo run tests and inspect coverage, run ``make test`` or ``make xtest`` for testing in parallel.\n\nTo rebuild in place after making changes, run ``make build``.\n\nTo create python package artifacts, run ``make dist``.\n\nTo build and view documentation, run ``make docs``.\n",
    'author': 'flynn',
    'author_email': 'crf204@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/crflynn/skgrf',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



from libcpp.memory cimport bool
from libcpp.memory cimport unique_ptr
from libcpp.set cimport set
from libcpp.unordered_map cimport unordered_map
from libcpp.vector cimport vector


cdef extern from "./grf/src/commons/Data.h" namespace "grf":
    cdef cppclass Data:
        Data(
            const double* data_ptr,
            size_t num_rows,
            size_t num_cols,
        ) except +
        void set_outcome_index(size_t index)
        void set_treatment_index(size_t index)
        void set_instrument_index(size_t index)
        void set_weight_index(size_t index)
        void set_causal_survival_numerator_index(size_t index)
        void set_causal_survival_denominator_index(size_t index)
        void set_censor_index(size_t index)
        vector[size_t] get_all_values(
            vector[double]& all_values,
            vector[size_t]& sorted_samples,
            const vector[size_t]& samples,
            size_t var
        ) const
        size_t get_num_cols() const
        size_t get_num_rows() const
        size_t get_num_outcomes() const
        size_t get_num_treatments() const
        const set[size_t]& get_disallowed_split_variables const
        double get_outcome(size_t row) const
        double get_treatment(size_t row) const
        double get_instrument(size_t row) const
        double get_weight(size_t row) const
        double get_causal_survival_numerator(size_t row) const
        double get_causal_survival_denominator(size_t row) const
        bool is_failure(size_t row) const
        double get(size_t row, size_t col) const


cdef extern from "./grf/src/commons/utility.h" namespace "grf":
    void split_sequence(
        vector[unsigned int]& result,
        unsigned int start,
        unsigned int end,
        unsigned int num_parts,
    )


cdef extern from "./grf/src/forest/Forest.h" namespace "grf":
    cdef cppclass Forest:
        Forest(
            vector[unique_ptr[Tree]]& trees,
            size_t num_variables,
            size_t ci_group_size,
        )
        Forest(Forest&& forest) except +
        const size_t get_ci_group_size() const
        const size_t get_num_variables() const
        const vector[unique_ptr[Tree]]& get_trees() const
        vector[unique_ptr[Tree]]& get_trees_()


cdef extern from "./grf/src/forest/ForestOptions.h" namespace "grf":
    cdef cppclass ForestOptions:
        ForestOptions() except +
        ForestOptions(
            unsigned int num_trees,
            size_t ci_group_size,
            double sample_fraction,
            unsigned int mtry,
            unsigned int min_node_size,
            bool honesty,
            double honesty_fraction,
            bool honesty_prune_leaves,
            double alpha,
            double imbalance_penalty,
            unsigned int num_threads,
            unsigned int random_seed,
            const vector[size_t]& sample_clusters,
            unsigned int samples_per_cluster,
        )


cdef extern from "./grf/src/prediction/collector/DefaultPredictionCollector.h" namespace "grf":
    cdef cppclass DefaultPredictionCollector:
        DefaultPredictionCollector() except +


cdef extern from "./grf/src/prediction/collector/OptimizedPredictionCollector.h" namespace "grf":
    cdef cppclass OptimizedPredictionCollector:
        OptimizedPredictionCollector() except +


cdef extern from "./grf/src/prediction/collector/SampleWeightComputer.h" namespace "grf":
    cdef cppclass SampleWeightComputer:
        SampleWeightComputer() except +
        unordered_map[size_t, double] compute_weights(
            size_t sample,
            const Forest& forest,
            const vector[vector[size_t]]& leaf_nodes_by_tree,
            const vector[vector[bool]]& valid_trees_by_sample,
        ) const


cdef extern from "./grf/src/prediction/collector/TreeTraverser.h" namespace "grf":
    cdef cppclass TreeTraverser:
        TreeTraverser(unsigned int num_threads) except +
        vector[vector[size_t]] get_leaf_nodes(
            const Forest& forest,
            const Data& data,
            bool oob_prediction
        ) const
        vector[vector[bool]] get_valid_trees_by_sample(
            const Forest& forest,
            const Data& data,
            bool oob_prediction
        ) const


cdef extern from "./grf/src/prediction/DefaultPredictionStrategy.h" namespace "grf":
    cdef cppclass DefaultPredictionStrategy:
        DefaultPredictionStrategy() except +


cdef extern from "./grf/src/prediction/CausalSurvivalPredictionStrategy.h" namespace "grf":
    cdef cppclass CausalSurvivalPredictionStrategy:
        CausalSurvivalPredictionStrategy() except +


cdef extern from "./grf/src/prediction/ProbabilityPredictionStrategy.h" namespace "grf":
    cdef cppclass ProbabilityPredictionStrategy:
        ProbabilityPredictionStrategy() except +


cdef extern from "./grf/src/prediction/MultiRegressionPredictionStrategy.h" namespace "grf":
    cdef cppclass MultiRegressionPredictionStrategy:
        MultiRegressionPredictionStrategy() except +


cdef extern from "./grf/src/prediction/InstrumentalPredictionStrategy.h" namespace "grf":
    cdef cppclass InstrumentalPredictionStrategy:
        InstrumentalPredictionStrategy() except +


cdef extern from "./grf/src/prediction/LLCausalPredictionStrategy.h" namespace "grf":
    cdef cppclass LLCausalPredictionStrategy:
        LLCausalPredictionStrategy() except +


cdef extern from "./grf/src/prediction/MultiCausalPredictionStrategy.h" namespace "grf":
    cdef cppclass MultiCausalPredictionStrategy:
        MultiCausalPredictionStrategy() except +


cdef extern from "./grf/src/prediction/LocalLinearPredictionStrategy.h" namespace "grf":
    cdef cppclass LocalLinearPredictionStrategy:
        LocalLinearPredictionStrategy() except +


cdef extern from "./grf/src/prediction/ObjectiveBayesDebiaser.h" namespace "grf":
    cdef cppclass ObjectiveBayesDebiaser:
        ObjectiveBayesDebiaser() except +


cdef extern from "./grf/src/prediction/Prediction.h" namespace "grf":
    cdef cppclass Prediction:
        Prediction(const vector[double]& predictions)
        Prediction(
            const vector[double]& predictions,
            const vector[double]& variance_estimates,
            const vector[double]& error_estimates,
            const vector[double]& excess_error_estimates,
        )
        const vector[double]& get_predictions() const
        const vector[double]& get_variance_estimates() const
        const vector[double]& get_error_estimates() const
        const vector[double]& get_excess_error_estimates() const
        const bool contains_variance_estimates() const
        const bool contains_error_estimates() const
        const size_t size() const


cdef extern from "./grf/src/prediction/PredictionValues.h" namespace "grf":
    cdef cppclass PredictionValues:
        PredictionValues() except +
        PredictionValues(
            const vector[vector[double]]& values,
            size_t num_types,
        )
        const vector[vector[double]]& get_all_values() const
        const size_t get_num_types() const


cdef extern from "./grf/src/prediction/OptimizedPredictionStrategy.h" namespace "grf":
    cdef cppclass OptimizedPredictionStrategy:
        OptimizedPredictionStrategy() except +


cdef extern from "./grf/src/prediction/QuantilePredictionStrategy.h" namespace "grf":
    cdef cppclass QuantilePredictionStrategy:
        QuantilePredictionStrategy() except +


cdef extern from "./grf/src/prediction/RegressionPredictionStrategy.h" namespace "grf":
    cdef cppclass RegressionPredictionStrategy:
        RegressionPredictionStrategy() except +


cdef extern from "./grf/src/prediction/SurvivalPredictionStrategy.h" namespace "grf":
    cdef cppclass SurvivalPredictionStrategy:
        SurvivalPredictionStrategy() except +


cdef extern from "./grf/src/relabeling/InstrumentalRelabelingStrategy.h" namespace "grf":
    cdef cppclass InstrumentalRelabelingStrategy:
        InstrumentalRelabelingStrategy() except +


cdef extern from "./grf/src/relabeling/CausalSurvivalRelabelingStrategy.h" namespace "grf":
    cdef cppclass CausalSurvivalRelabelingStrategy:
        CausalSurvivalRelabelingStrategy() except +


cdef extern from "./grf/src/relabeling/MultiCausalRelabelingStrategy.h" namespace "grf":
    cdef cppclass MultiCausalRelabelingStrategy:
        MultiCausalRelabelingStrategy() except +


cdef extern from "./grf/src/relabeling/LLRegressionRelabelingStrategy.h" namespace "grf":
    cdef cppclass LLRegressionRelabelingStrategy:
        LLRegressionRelabelingStrategy() except +


cdef extern from "./grf/src/relabeling/NoopRelabelingStrategy.h" namespace "grf":
    cdef cppclass NoopRelabelingStrategy:
        NoopRelabelingStrategy() except +


cdef extern from "./grf/src/relabeling/MultiNoopRelabelingStrategy.h" namespace "grf":
    cdef cppclass MultiNoopRelabelingStrategy:
        MultiNoopRelabelingStrategy() except +


cdef extern from "./grf/src/relabeling/QuantileRelabelingStrategy.h" namespace "grf":
    cdef cppclass QuantileRelabelingStrategy:
        QuantileRelabelingStrategy() except +


cdef extern from "./grf/src/relabeling/RelabelingStrategy.h" namespace "grf":
    cdef cppclass RelabelingStrategy:
        RelabelingStrategy() except +


cdef extern from "./grf/src/sampling/RandomSampler.h" namespace "grf":
    cdef cppclass RandomSampler:
        RandomSampler() except +


cdef extern from "./grf/src/sampling/SamplingOptions.h" namespace "grf":
    cdef cppclass SamplingOptions:
        SamplingOptions() except +


cdef extern from "./grf/src/splitting/factory/InstrumentalSplittingRuleFactory.h" namespace "grf":
    cdef cppclass InstrumentalSplittingRuleFactory:
        InstrumentalSplittingRuleFactory() except +


cdef extern from "./grf/src/splitting/factory/CausalSurvivalSplittingRuleFactory.h" namespace "grf":
    cdef cppclass CausalSurvivalSplittingRuleFactory:
        CausalSurvivalSplittingRuleFactory() except +


cdef extern from "./grf/src/splitting/factory/MultiCausalSplittingRuleFactory.h" namespace "grf":
    cdef cppclass MultiCausalSplittingRuleFactory:
        MultiCausalSplittingRuleFactory() except +


cdef extern from "./grf/src/splitting/factory/MultiRegressionSplittingRuleFactory.h" namespace "grf":
    cdef cppclass MultiRegressionSplittingRuleFactory:
        MultiRegressionSplittingRuleFactory() except +


cdef extern from "./grf/src/splitting/factory/RegressionSplittingRuleFactory.h" namespace "grf":
    cdef cppclass RegressionSplittingRuleFactory:
        RegressionSplittingRuleFactory() except +


cdef extern from "./grf/src/splitting/factory/ProbabilitySplittingRuleFactory.h" namespace "grf":
    cdef cppclass ProbabilitySplittingRuleFactory:
        ProbabilitySplittingRuleFactory() except +


cdef extern from "./grf/src/splitting/factory/SplittingRuleFactory.h" namespace "grf":
    cdef cppclass SplittingRuleFactory:
        SplittingRuleFactory() except +


cdef extern from "./grf/src/splitting/factory/SurvivalSplittingRuleFactory.h" namespace "grf":
    cdef cppclass SurvivalSplittingRuleFactory:
        SurvivalSplittingRuleFactory() except +


cdef extern from "./grf/src/splitting/InstrumentalSplittingRule.h" namespace "grf":
    cdef cppclass InstrumentalSplittingRule:
        InstrumentalSplittingRule() except +


cdef extern from "./grf/src/splitting/CausalSurvivalSplittingRule.h" namespace "grf":
    cdef cppclass CausalSurvivalSplittingRule:
        CausalSurvivalSplittingRule() except +


cdef extern from "./grf/src/splitting/MultiCausalSplittingRule.h" namespace "grf":
    cdef cppclass MultiCausalSplittingRule:
        MultiCausalSplittingRule() except +


cdef extern from "./grf/src/splitting/MultiRegressionSplittingRule.h" namespace "grf":
    cdef cppclass MultiRegressionSplittingRule:
        MultiRegressionSplittingRule() except +


cdef extern from "./grf/src/splitting/ProbabilitySplittingRule.h" namespace "grf":
    cdef cppclass ProbabilitySplittingRule:
        ProbabilitySplittingRule() except +


cdef extern from "./grf/src/splitting/RegressionSplittingRule.h" namespace "grf":
    cdef cppclass RegressionSplittingRule:
        RegressionSplittingRule() except +


cdef extern from "./grf/src/splitting/SurvivalSplittingRule.h" namespace "grf":
    cdef cppclass SurvivalSplittingRule:
        SurvivalSplittingRule() except +


cdef extern from "./grf/src/tree/Tree.h" namespace "grf":
    cdef cppclass Tree:
        Tree(
            size_t root_node,
            const vector[vector[size_t]]& child_nodes,
            const vector[vector[size_t]]& leaf_samples,
            const vector[size_t]& split_vars,
            const vector[double]& split_values,
            const vector[size_t]& drawn_samples,
            const vector[bool]& send_missing_left,
            const PredictionValues& prediction_values,
        )
        size_t get_root_node() const
        const vector[vector[size_t]]& get_child_nodes() const
        const vector[vector[size_t]]& get_leaf_samples() const
        const vector[size_t]& get_split_vars() const
        const vector[double]& get_split_values() const
        const vector[size_t]& get_drawn_samples() const
        const vector[bool]& get_send_missing_left() const
        const PredictionValues& get_prediction_values() const


cdef extern from "./grf/src/tree/TreeOptions.h" namespace "grf":
    cdef cppclass TreeOptions:
        TreeOptions() except +


cdef extern from "./grf/src/tree/TreeTrainer.h" namespace "grf":
    cdef cppclass TreeTrainer:
        TreeTrainer() except +


cdef extern from "./grf/src/forest/ForestPredictor.h" namespace "grf":
    cdef cppclass ForestPredictor:
        ForestPredictor(ForestPredictor&)
        ForestPredictor(
            unsigned int num_threads,
            unique_ptr[DefaultPredictionStrategy] strategy,
        )
        ForestPredictor(
            unsigned int num_threads,
            unique_ptr[OptimizedPredictionStrategy] strategy,
        )
        vector[Prediction] predict(
            const Forest& forest,
            const Data& train_data,
            const Data& data,
            bool estimate_variance
        ) const
        vector[Prediction] predict_oob(
            const Forest& forest,
            const Data& data,
            bool estimate_variance,
        ) const


cdef extern from "./grf/src/forest/ForestPredictors.h" namespace "grf":
    cdef ForestPredictor instrumental_predictor(unsigned int num_threads)
    cdef ForestPredictor ll_regression_predictor(
        unsigned int num_threads,
        vector[double] lambdas,
        bool weight_penalty,
        vector[size_t] linear_correction_variables,
    )
    cdef ForestPredictor probability_predictor(
        unsigned int num_threads,
        size_t num_classes,
    )
    cdef ForestPredictor quantile_predictor(
        unsigned int num_threads,
        const vector[double]& quantiles,
    )
    cdef ForestPredictor regression_predictor(unsigned int num_threads)
    cdef ForestPredictor survival_predictor(
        unsigned int num_threads,
        size_t num_failures,
        int prediction_type,
    )


cdef extern from "./grf/src/forest/ForestTrainer.h" namespace "grf":
    cdef cppclass ForestTrainer:
        # no nullary constructor
        # and we have to explicitly declare the copy constructor
        # so that we can instantiate with regression_trainer()
        ForestTrainer(ForestTrainer&)
        ForestTrainer(
            unique_ptr[RelabelingStrategy] relabeling_strategy,
            unique_ptr[SplittingRuleFactory] splitting_rule_factory,
            unique_ptr[OptimizedPredictionStrategy] prediction_strategy,
        )
        Forest train(
            const Data& data,
            const ForestOptions& options,
        )


cdef extern from "./grf/src/forest/ForestTrainers.h" namespace "grf":
    cdef ForestTrainer instrumental_trainer(
        double reduced_form_weight,
        bool stabilize_splits,
    )
    cdef ForestTrainer ll_regression_trainer(
        double split_lambda,
        bool weight_penalty,
        const vector[double]& overall_beta,
        size_t ll_split_cutoff,
        vector[size_t] ll_split_variables,
    )
    cdef ForestTrainer probability_trainer(size_t num_classes)
    cdef ForestTrainer quantile_trainer(const vector[double]& quantiles)
    cdef ForestTrainer regression_trainer()
    cdef ForestTrainer survival_trainer()


cdef extern from "./grf/src/analysis/SplitFrequencyComputer.h" namespace "grf":
    cdef cppclass SplitFrequencyComputer:
        vector[vector[size_t]] compute(
            const Forest& forest,
            size_t max_depth
        )


import typing as t

import logging

from skgrf.tree.instrumental_regressor import GRFTreeInstrumentalRegressor

if t.TYPE_CHECKING:  # pragma: no cover
    from skgrf.ensemble import GRFForestCausalRegressor

logger = logging.getLogger(__name__)


class GRFTreeCausalRegressor(GRFTreeInstrumentalRegressor):
    r"""GRF Tree Causal regression implementation for sci-kit learn.

    Provides a sklearn tree causal regressor to the GRF C++ library using Cython.

    :param bool equalize_cluster_weights: Weight the samples such that clusters have
        equally weight. If ``False``, larger clusters will have more weight. If
        ``True``, the number of samples drawn from each cluster is equal to the size of
        the smallest cluster. If ``True``, sample weights should not be passed on
        fitting.
    :param float sample_fraction: Fraction of samples used in each tree. If
        ``ci_group_size`` > 1, the max allowed fraction is 0.5
    :param int mtry: The number of features to split on each node. The default is
        ``sqrt(p) + 20`` where ``p`` is the number of features.
    :param int min_node_size: The minimum number of observations in each tree leaf.
    :param bool honesty: Use honest splitting (subsample splitting).
    :param float honesty_fraction: The fraction of data used for subsample splitting.
    :param bool honesty_prune_leaves: Prune estimation sample tree such that no leaves
        are empty. If ``False``, trees with empty leaves are skipped.
    :param float alpha: The maximum imbalance of a split.
    :param float imbalance_penalty: Penalty applied to imbalanced splits.
    :param bool orthogonal_boosting: When ``y_hat`` or ``w_hat`` are ``None``, they
        are estimated using boosted regression forests. (Not yet implemented)
    :param bool stabilize_splits: Whether or not the instrument should be taken into
        account when determining the imbalance of a split.
    :param int n_jobs: The number of threads. Default is number of CPU cores. Only used
        for target estimation.
    :param int seed: Random seed value.

    :ivar int n_features_in\_: The number of features (columns) from the fit input
        ``X``.
    :ivar dict grf_forest\_: The returned result object from calling C++ grf.
    :ivar int mtry\_: The ``mtry`` value determined by validation.
    :ivar int outcome_index\_: The index of the grf train matrix holding the outcomes.
    :ivar list samples_per_cluster\_: The number of samples to train per cluster.
    :ivar list clusters\_: The cluster labels determined from the fit input ``cluster``.
    :ivar int n_clusters\_: The number of unique cluster labels from the fit input
        ``cluster``.
    :ivar str criterion: The criterion used for splitting: ``mse``
    """

    def __init__(
        self,
        equalize_cluster_weights=False,
        sample_fraction=0.5,
        mtry=None,
        min_node_size=5,
        honesty=True,
        honesty_fraction=0.5,
        honesty_prune_leaves=True,
        alpha=0.05,
        imbalance_penalty=0,
        stabilize_splits=True,
        orthogonal_boosting=False,
        n_jobs=-1,
        seed=42,
    ):
        super().__init__(
            equalize_cluster_weights=equalize_cluster_weights,
            sample_fraction=sample_fraction,
            mtry=mtry,
            min_node_size=min_node_size,
            honesty=honesty,
            honesty_fraction=honesty_fraction,
            honesty_prune_leaves=honesty_prune_leaves,
            alpha=alpha,
            imbalance_penalty=imbalance_penalty,
            reduced_form_weight=0,
            stabilize_splits=stabilize_splits,
            n_jobs=n_jobs,
            seed=seed,
        )
        self.orthogonal_boosting = orthogonal_boosting

    @classmethod
    def from_forest(cls, forest: "GRFForestCausalRegressor", idx: int):
        """Extract a tree from a forest.

        :param GRFLocalLinearRegressor forest: A trained GRFLocalLinearRegressor
            instance
        :param int idx: The tree index from the forest to extract.
        """
        # Even though we have a tree object, we keep the exact same dictionary structure
        # that exists in the forests, so that we can reuse the Cython entrypoints.
        # We also copy over some instance attributes from the trained forest.

        # params
        instance = cls(
            equalize_cluster_weights=forest.equalize_cluster_weights,
            sample_fraction=forest.sample_fraction,
            mtry=forest.mtry,
            min_node_size=forest.min_node_size,
            honesty=forest.honesty,
            honesty_fraction=forest.honesty_fraction,
            honesty_prune_leaves=forest.honesty_prune_leaves,
            alpha=forest.alpha,
            imbalance_penalty=forest.imbalance_penalty,
            stabilize_splits=forest.stabilize_splits,
            n_jobs=forest.n_jobs,
            seed=forest.seed,
        )
        # forest
        grf_forest = {}
        for k, v in forest.grf_forest_.items():
            if isinstance(v, list):
                grf_forest[k] = [forest.grf_forest_[k][idx]]
            else:
                grf_forest[k] = v
        grf_forest["num_trees"] = 1
        instance.grf_forest_ = grf_forest
        instance._ensure_ptr()
        # vars
        instance.outcome_index_ = forest.outcome_index_
        instance.treatment_index_ = forest.treatment_index_
        instance.instrument_index_ = forest.instrument_index_
        instance.n_features_in_ = forest.n_features_in_
        instance.clusters_ = forest.clusters_
        instance.n_clusters_ = forest.n_clusters_
        instance.samples_per_cluster_ = forest.samples_per_cluster_
        instance.mtry_ = forest.mtry_
        instance.sample_weight_index_ = forest.sample_weight_index_
        return instance

    # noinspection PyMethodOverriding
    def fit(
        self,
        X,
        y,
        w,  # treatment
        y_hat=None,
        w_hat=None,
        sample_weight=None,
        cluster=None,
    ):
        """Fit the grf forest using training data.

        :param array2d X: training input features
        :param array1d y: training input targets
        :param array1d w: training input treatments
        :param array1d y_hat: estimated expected target responses
        :param array1d w_hat: estimated treatment propensities
        :param array1d sample_weight: optional weights for input samples
        :param array1d cluster: optional cluster assignments for input samples
        """
        # avoiding circular import
        from skgrf.ensemble.boosted_regressor import GRFBoostedForestRegressor

        X, y = self._validate_data(X, y)
        self._check_num_samples(X)

        boost_params = {
            "n_estimators": 500,
            "equalize_cluster_weights": self.equalize_cluster_weights,
            "sample_fraction": self.sample_fraction,
            "mtry": self.mtry,
            "min_node_size": 5,
            "honesty": True,
            "honesty_fraction": 0.5,
            "honesty_prune_leaves": self.honesty_prune_leaves,
            "alpha": self.alpha,
            "imbalance_penalty": self.imbalance_penalty,
            "ci_group_size": 1,
            "tune_params": None,  # TODO ?
            "n_jobs": self.n_jobs,
            "seed": self.seed,
        }
        if y_hat is None and self.orthogonal_boosting:
            logger.debug("orthogonal boosting y_hat")
            br = GRFBoostedForestRegressor(**boost_params)
            br.fit(X, y, sample_weight=sample_weight, cluster=cluster)
            y_hat = br.boosted_forests_["predictions"]

        if w_hat is None and self.orthogonal_boosting:
            logger.debug("orthogonal boosting w_hat")
            br = GRFBoostedForestRegressor(**boost_params)
            br.fit(X, w, sample_weight=sample_weight, cluster=cluster)
            w_hat = br.boosted_forests_["predictions"]

        return super().fit(
            X=X,
            y=y,
            w=w,
            z=w,
            y_hat=y_hat,
            w_hat=w_hat,
            z_hat=w_hat,
            cluster=cluster,
            sample_weight=sample_weight,
        )
